# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import netaddr
from oslo_log import log
import typing  # noqa

import tobiko
from tobiko.shell import sh
from tobiko.shell import ssh


class SocketLookupError(tobiko.TobikoException):
    message = ('ss command "{cmd}" failed with the error "{err}"')


class SockHeader():

    def __init__(self, header_str: str):
        self.header_str = header_str
        self.header: typing.List[str] = []
        self._parse_header()

    def _parse_header(self):
        if 'Netid' in self.header_str:
            self.header.append('protocol')
        if 'State' in self.header_str:
            self.header.append('state')
        if 'Recv-Q' in self.header_str:
            self.header.append('recv_q')
        if 'Send-Q' in self.header_str:
            self.header.append('send_q')
        if 'Local Address:Port' in self.header_str:
            self.header.append('local')
        if 'Peer Address:Port' in self.header_str:
            self.header.append('remote')
        if 'Process' in self.header_str:
            self.header.append('process')

    def extend_ports(self):
        """Add ports as separate header columns

        It might be useful while collecting information about unix sockets
        as unlke regular network sockets there are no ':' symbol between
        address and port. But the ' ' (space) symbol is used instead.
        """
        try:
            idx = self.header.index('local')
            self.header.pop(idx)
            self.header.insert(idx, 'local_addr')
            self.header.insert(idx+1, 'local_port')
        except ValueError:
            pass
        try:
            idx = self.header.index('remote')
            self.header.pop(idx)
            self.header.insert(idx, 'remote_addr')
            self.header.insert(idx+1, 'remote_port')
        except ValueError:
            pass

    def __len__(self):
        return len(self.header)

    def __iter__(self):
        yield from self.header


class SockLine(str):
    """Single line from the output of ss command line tool

    It should match with the corresponding table header that is presented by
    object of SockHeader class
    """


class SockData(dict):
    """A single socket information parsed from output of ss command line tool

    Output of ss command line tool should be parsed and stored in the dict
    with keys are items of the SockHeader object. In most of the cases it
    should contain the following keys:

    - protocol (optional)
    - state (optional)
    - recv_q
    - send_q
    - local_addr (IP or filename)
    - local_port
    - remote_addr (IP or filename)
    - remote_port
    - process (list of processes names)
    """


LOG = log.getLogger(__name__)


def _ss(params: str = '',
        ssh_client: ssh.SSHClientFixture = None,
        table_header: SockHeader = None,
        parser: typing.Callable[[SockHeader, SockLine], SockData] = None,
        **execute_params) -> typing.List[SockData]:
    execute_params.update({'sudo': True})
    sockets = []
    headers = None
    if table_header:
        # Predefined header might be necessary if the command is expected to
        # be executed in any kind of environments. Old versrions of `ss`
        # command line tool do not have 'Processes' column in header
        parsed_header = True
        headers = table_header
        command_line = "ss -Hnp {}".format(params)
    else:
        parsed_header = False
        command_line = "ss -np {}".format(params)
    try:
        stdout = sh.execute(command_line,
                            ssh_client=ssh_client,
                            **execute_params).stdout
    except sh.ShellCommandFailed as ex:
        if ex.stdout.startswith('Error'):
            raise SocketLookupError(cmd=command_line, err=ex.stderr) from ex
        if ex.exit_status > 0:
            raise
    for line in stdout.splitlines():
        if not parsed_header:
            headers = SockHeader(line)
            parsed_header = True
            continue
        sock_info = SockLine(line.strip())
        if parser and headers:
            try:
                sockets.append(parser(headers, sock_info))
            except ValueError as ex:
                LOG.error(str(ex))
                continue
        else:
            sockets.append(SockData({'raw_data': sock_info}))
    return sockets


def get_processes(processes: str) -> typing.List[str]:
    """Parse processes names from ss output

    The simpliest example of the proccesses suffix in ss output:

        users:(("httpd",pid=735448,fd=11))

    But it can be a bit more complex

        users:(("httpd",pid=4969,fd=53),("httpd",pid=3328,fd=53))

    Function return the list of all processes names ['httpd', 'httpd']
    """
    stack = []
    process_list = []
    nested = False
    for idx, symbol in enumerate(processes):
        if symbol == '(':
            stack.append(idx)
            nested = True
        elif symbol == ')' and len(stack) == 1:
            process_list.extend(get_processes(processes[stack[0]+1:idx]))
        elif symbol == ')':
            stack.pop()
    if not nested:
        process_list.append(processes.split('"', 2)[1])
    return process_list


def parse_tcp_socket(headers: SockHeader,
                     sock_info: SockLine) -> SockData:
    socket_details = SockData()
    sock_data = sock_info.split()
    if len(headers) != len(sock_data):
        msg = 'Unable to parse line: "{}"'.format(sock_info)
        raise ValueError(msg)
    for idx, header in enumerate(headers):
        if not header:
            continue
        if header == 'local' or header == 'remote':
            ip, port = sock_data[idx].strip().rsplit(':', 1)
            if ip == '*':
                ip = '0'
            socket_details['{}_addr'.format(header)] = netaddr.IPAddress(
                    ip.strip(']['))
            socket_details['{}_port'.format(header)] = port
        elif header == 'process':
            try:
                socket_details[header] = get_processes(sock_data[idx])
            except IndexError as ex:
                msg = 'Unable to parse processes part of the line: {}'.format(
                        sock_info)
                raise ValueError(msg) from ex
        else:
            socket_details[header] = sock_data[idx]
    return socket_details


def parse_unix_socket(headers: SockHeader,
                      sock_info: SockLine) -> SockData:
    socket_details = SockData()
    sock_data = sock_info.split()
    if len(headers) != len(sock_data):
        msg = 'Unable to parse line: "{}"'.format(sock_info)
        raise ValueError(msg)
    for idx, header in enumerate(headers):
        if not header:
            continue
        if header == 'process':
            try:
                socket_details[header] = get_processes(sock_data[idx])
            except IndexError as ex:
                msg = 'Unable to parse processes part of the line: {}'.format(
                        sock_info)
                raise ValueError(msg) from ex
        else:
            socket_details[header] = sock_data[idx]
    return socket_details


def tcp_listening(address: str = '',
                  port: str = '',
                  **exec_params) -> typing.List[SockData]:
    """List of tcp sockets in listening state

    Information can be filtered by local address and port and will be returned
    as a list of SockData object where the following keys should exist:

    - protocol (optional)
    - state (optional)
    - recv_q
    - send_q
    - local_addr (netaddr.IPAddress object)
    - local_port
    - remote_addr (netaddr.IPAddress object)
    - remote_port
    - process (list of processes names)
    """
    params = '-t state listening'
    ss_fields = SockHeader('Recv-Q Send-Q Local Address:Port'
                           'Peer Address:Port Process')
    if port:
        params += ' sport {}'.format(port)
    if address:
        params += ' src {}'.format(address)
    return _ss(params=params,
               table_header=ss_fields,
               parser=parse_tcp_socket,
               **exec_params)


def tcp_connected(src_address: str = '',
                  src_port: str = '',
                  dst_address: str = '',
                  dst_port: str = '',
                  **exec_params) -> typing.List[SockData]:
    params = '-t state connected'
    ss_fields = SockHeader('State Recv-Q Send-Q Local Address:Port'
                           'Peer Address:Port Process')
    if src_port:
        params += ' sport {}'.format(src_port)
    if src_address:
        params += ' src {}'.format(src_address)
    if dst_port:
        params += ' dport {}'.format(dst_port)
    if dst_address:
        params += ' dst {}'.format(dst_address)
    return _ss(params=params,
               table_header=ss_fields,
               parser=parse_tcp_socket,
               **exec_params)


def unix_listening(file_name: str = '',
                   **exec_params) -> typing.List[SockData]:
    """List of unix sockets in listening state

    Information can be filtered by file name and will be returned as a list of
    SockData object where the following keys should exist:

    - protocol (optional)
    - state (optional)
    - recv_q
    - send_q
    - local_addr (filename string)
    - local_port (should be ``*``)
    - remote_addr (filename string - should be ``*``)
    - remote_port (should be ``*``)
    - process (list of processes names)
    """
    params = '-x state listening'
    ss_fields = SockHeader('Netid Recv-Q Send-Q Local Address:Port'
                           'Peer Address:Port Process')
    ss_fields.extend_ports()
    if file_name:
        params += ' src {}'.format(file_name)
    return _ss(params=params,
               table_header=ss_fields,
               parser=parse_unix_socket,
               **exec_params)



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
from __future__ import absolute_import

import testtools


from tobiko.tests.faults.podified.ha import cloud_disruptions
from tobiko import podified


@podified.skip_if_not_podified
class DisruptPodifiedNodesTest(testtools.TestCase):
    """ HA Tests: run health check -> disruptive action -> health check
    disruptive_action: a function that runs some
    disruptive scenario on a node"""

    def test_kill_all_galera_services(self):
        # HealthCheck.run_before()
        cloud_disruptions.kill_all_galera_services()
        # HealthCheck.run_after()



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os

import testtools

import tobiko
from tobiko import config
from tobiko import tripleo


CONF = config.CONF
PLAYBOOK_DIRNAME = os.path.join(os.path.dirname(__file__), 'playbooks')


@tripleo.skip_if_missing_undercloud
class InventoryFileTest(testtools.TestCase):

    @tripleo.skip_if_missing_tripleo_ansible_inventory
    def test_get_tripleo_ansible_inventory(self):
        inventory = tripleo.get_tripleo_ansible_inventory()
        self.assertIn('Undercloud', inventory)
        self.assertIn('Controller', inventory)
        self.assertIn('Compute', inventory)

    @tripleo.skip_if_missing_tripleo_ansible_inventory
    def test_get_tripleo_ansible_inventory_file(self):
        inventory_file = tripleo.get_tripleo_ansible_inventory_file()
        self.assertTrue(os.path.isfile(inventory_file))

    def test_has_tripleo_ansible_inventory(self):
        result = tripleo.has_tripleo_ansible_inventory()
        inventory_file = tripleo.get_tripleo_ansible_inventory_file()
        self.assertIs(inventory_file and os.path.isfile(inventory_file),
                      result)

    def test_read_tripleo_ansible_inventory(self):
        inventory_yaml = tripleo.read_tripleo_ansible_inventory()
        self.assertIsInstance(inventory_yaml, str)
        self.assertTrue(inventory_yaml)
        inventory = tobiko.load_yaml(inventory_yaml)
        self.assertIn('Undercloud', inventory)
        self.assertIn('Controller', inventory)
        self.assertIn('Compute', inventory)


@tripleo.skip_if_missing_tripleo_ansible_inventory
class PlaybookTest(testtools.TestCase):

    def test_ping_hosts(self):
        tripleo.run_playbook_from_undercloud(
            playbook='test_ping_hosts.yaml',
            playbook_dirname=PLAYBOOK_DIRNAME,
            roles=['ping'])

    def test_debug_vars(self):
        tripleo.run_playbook_from_undercloud(
            playbook='test_debug_vars.yaml',
            playbook_dirname=PLAYBOOK_DIRNAME,
            playbook_files=['vars/some-vars.yaml'])

    def test_undecloud_current_dir(self):
        tripleo.run_playbook_from_undercloud(
            playbook='test_undercloud_current_dir.yaml',
            playbook_dirname=PLAYBOOK_DIRNAME)

    @tripleo.skip_unless_undercloud_has_ansible(min_version=2.9)
    def test_overcloud_openstack_auth(self):
        tripleo.run_playbook_from_undercloud(
            playbook='test_overcloud_openstack_auth.yaml',
            playbook_dirname=PLAYBOOK_DIRNAME,
            requirements_files=['requirements/openstack-cloud.yaml'])



"""Models for manipulating containers and storage."""

from __future__ import absolute_import

import collections
import getpass
import json
import logging
import signal
import time

from . import fold_keys
from ._containers_attach import Mixin as AttachMixin
from ._containers_start import Mixin as StartMixin


class Container(AttachMixin, StartMixin, collections.UserDict):
    """Model for a container."""

    def __init__(self, client, ident, data, refresh=True):
        """Construct Container Model."""
        super(Container, self).__init__(data)
        self._client = client
        self._id = ident

        if refresh:
            with client() as podman:
                self._refresh(podman)
        else:
            for k, v in self.data.items():
                setattr(self, k, v)
            if 'containerrunning' in self.data:
                setattr(self, 'running', self.data['containerrunning'])
                self.data['running'] = self.data['containerrunning']

        assert self._id == data['id'],\
            'Requested container id({}) does not match store id({})'.format(
                self._id, data['id']
            )

    def _refresh(self, podman, tries=1):
        try:
            ctnr = podman.GetContainer(self._id)
        except BrokenPipeError:
            logging.debug('Failed GetContainer(%s) try %d/3', self._id, tries)
            if tries > 3:
                raise
            else:
                with self._client() as pman:
                    self._refresh(pman, tries + 1)
        else:
            super().update(ctnr['container'])

            for k, v in self.data.items():
                setattr(self, k, v)
            if 'containerrunning' in self.data:
                setattr(self, 'running', self.data['containerrunning'])
                self.data['running'] = self.data['containerrunning']

            return self

    def refresh(self):
        """Refresh status fields for this container."""
        with self._client() as podman:
            return self._refresh(podman)

    def processes(self):
        """Show processes running in container."""
        with self._client() as podman:
            results = podman.ListContainerProcesses(self._id)
        yield from results['container']

    def changes(self):
        """Retrieve container changes."""
        with self._client() as podman:
            results = podman.ListContainerChanges(self._id)
        return results['container']

    def kill(self, sig=signal.SIGTERM, wait=25):
        """Send signal to container.

        default signal is signal.SIGTERM.
        wait n of seconds, 0 waits forever.
        """
        with self._client() as podman:
            podman.KillContainer(self._id, sig)
            timeout = time.time() + wait
            while True:
                self._refresh(podman)
                if self.status != 'running':  # pylint: disable=no-member
                    return self

                if wait and timeout < time.time():
                    raise TimeoutError()

                time.sleep(0.5)

    def inspect(self):
        """Retrieve details about containers."""
        with self._client() as podman:
            results = podman.InspectContainer(self._id)
        obj = json.loads(results['container'], object_hook=fold_keys())
        return collections.namedtuple('ContainerInspect', obj.keys())(**obj)

    def export(self, target):
        """Export container from store to tarball.

        TODO: should there be a compress option, like images?
        """
        with self._client() as podman:
            results = podman.ExportContainer(self._id, target)
        return results['tarfile']

    def commit(self, image_name, **kwargs):
        """Create image from container.

        Keyword arguments:
            author -- change image's author
            message -- change image's message, docker format only.
            pause -- pause container during commit
            change -- Additional properties to change

        Change examples:
            CMD=/usr/bin/zsh
            ENTRYPOINT=/bin/sh date
            ENV=TEST=test_containers.TestContainers.test_commit
            EXPOSE=8888/tcp
            LABEL=unittest=test_commit
            USER=bozo:circus
            VOLUME=/data
            WORKDIR=/data/application

        All changes overwrite existing values.
          See inspect() to obtain current settings.
        """
        author = kwargs.get('author', None) or getpass.getuser()
        change = kwargs.get('change', None) or []
        message = kwargs.get('message', None) or ''
        pause = kwargs.get('pause', None) or True

        for c in change:
            if c.startswith('LABEL=') and c.count('=') < 2:
                raise ValueError(
                    'LABEL should have the format: LABEL=label=value, not {}'.
                    format(c))

        with self._client() as podman:
            results = podman.Commit(self._id, image_name, change, author,
                                    message, pause)
        return results['reply']['id']

    def stop(self, timeout=25):
        """Stop container, return id on success."""
        with self._client() as podman:
            podman.StopContainer(self._id, timeout)
            return self._refresh(podman)

    def remove(self, force=False):
        """Remove container, return id on success.

        force=True, stop running container.
        """
        with self._client() as podman:
            results = podman.RemoveContainer(self._id, force)
        return results['container']

    def restart(self, timeout=25):
        """Restart container with timeout, return id on success."""
        with self._client() as podman:
            podman.RestartContainer(self._id, timeout)
            return self._refresh(podman)

    def pause(self):
        """Pause container, return id on success."""
        with self._client() as podman:
            podman.PauseContainer(self._id)
            return self._refresh(podman)

    def unpause(self):
        """Unpause container, return id on success."""
        with self._client() as podman:
            podman.UnpauseContainer(self._id)
            return self._refresh(podman)

    def update_container(self, *args, **kwargs):  \
            # pylint: disable=unused-argument
        """TODO: Update container..., return id on success."""
        with self._client() as podman:
            podman.UpdateContainer()
            return self._refresh(podman)

    def wait(self):
        """Wait for container to finish, return 'returncode'."""
        with self._client() as podman:
            results = podman.WaitContainer(self._id)
        return int(results['exitcode'])

    def stats(self):
        """Retrieve resource stats from the container."""
        with self._client() as podman:
            results = podman.GetContainerStats(self._id)
        obj = results['container']
        return collections.namedtuple('StatDetail', obj.keys())(**obj)

    def logs(self, *args, **kwargs):  # pylint: disable=unused-argument
        """Retrieve container logs."""
        with self._client() as podman:
            results = podman.GetContainerLogs(self._id)
        yield from results['container']


class Containers():
    """Model for Containers collection."""

    def __init__(self, client):
        """Construct model for Containers collection."""
        self._client = client

    def list(self):
        """List of containers in the container store."""
        with self._client() as podman:
            results = podman.ListContainers()
        for cntr in results['containers']:
            yield Container(self._client, cntr['id'], cntr, refresh=False)

    def delete_stopped(self):
        """Delete all stopped containers."""
        with self._client() as podman:
            results = podman.DeleteStoppedContainers()
        return results['containers']

    def get(self, id_):
        """Retrieve container details from store."""
        with self._client() as podman:
            cntr = podman.GetContainer(id_)
        return Container(self._client, cntr['container']['id'],
                         cntr['container'])



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.metalsmith import _client
from tobiko.openstack.metalsmith import _instance


CLIENT_CLASSES = _client.CLIENT_CLASSES
MetalsmithClient = _client.MetalsmithClient
MetalsmithClientFixture = _client.MetalsmithClientFixture
MetalsmithClientType = _client.MetalsmithClientType
metalsmith_client = _client.metalsmith_client
get_metalsmith_client = _client.get_metalsmith_client

MetalsmithInstance = _instance.MetalsmithInstance
list_instances = _instance.list_instances
find_instance = _instance.find_instance
list_instance_ip_addresses = _instance.list_instance_ip_addresses
find_instance_ip_address = _instance.find_instance_ip_address



# Copyright 2023 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

from oslo_log import log

import tobiko
from tobiko.openstack import neutron
from tobiko.openstack import topology
from tobiko.podified import _edpm
from tobiko.podified import _openshift
from tobiko.podified import containers
from tobiko import rhosp
from tobiko.shell import sh
from tobiko.shell import ssh

LOG = log.getLogger(__name__)


skip_if_not_podified = tobiko.skip_unless(
    "Podified deployment not configured", _openshift.has_podified_cp
)
skip_if_podified = tobiko.skip_if(
    "This test cannot run on a Podified deployment", _openshift.has_podified_cp
)

# In Podified topology there are groups like 'edpm-compute', 'edpm-networker'
# and 'edpm-other' but we need to provide also "virtual" group which will
# contain all of those "sub groups"
COMPUTE_GROUPS = [
    _openshift.EDPM_COMPUTE_GROUP,
    _openshift.EDPM_NETWORKER_GROUP,
    _openshift.EDPM_OTHER_GROUP
]
ALL_COMPUTES_GROUP_NAME = 'compute'
OCP_WORKER = 'ocp_worker'
EDPM_NODE = 'edpm_node'


class PodifiedTopology(rhosp.RhospTopology):

    # NOTE(slaweq): those service names are only valid for the EDPM nodes
    agent_to_service_name_mappings = {
        neutron.DHCP_AGENT: 'edpm_neutron_dhcp',
        neutron.L3_AGENT:  'edpm_neutron_l3_agent',
        neutron.OPENVSWITCH_AGENT: 'edpm_neutron_ovs_agent',
        neutron.METADATA_AGENT: 'edpm_neutron_metadata_agent',
        neutron.OVN_METADATA_AGENT: 'edpm_ovn_metadata_agent',
        neutron.NEUTRON_OVN_METADATA_AGENT: 'edpm_ovn_metadata_agent',
        neutron.OVN_CONTROLLER: 'edpm_ovn_controller',
        neutron.OVN_BGP_AGENT: 'edpm_ovn_bgp_agent',
        neutron.FRR: 'edpm_frr'
    }

    # NOTE(slaweq): those container names are only valid for the EDPM nodes
    agent_to_container_name_mappings = {
        neutron.DHCP_AGENT: 'neutron_dhcp',
        neutron.L3_AGENT:  'neutron_l3_agent',
        neutron.OPENVSWITCH_AGENT: 'neutron_ovs_agent',
        neutron.METADATA_AGENT: 'neutron_metadata_agent',
        neutron.OVN_METADATA_AGENT: 'ovn_metadata_agent',
        neutron.NEUTRON_OVN_METADATA_AGENT: 'ovn_metadata_agent',
        neutron.OVN_CONTROLLER: 'ovn_controller',
        neutron.OVN_BGP_AGENT: 'ovn_bgp_agent',
        neutron.FRR: 'frr'
    }

    sidecar_container_list = [
        'neutron-haproxy-ovnmeta',
        'neutron-dnsmasq-qdhcp'
    ]

    @property
    def ignore_containers_list(self):
        return self.sidecar_container_list

    def assert_containers_running(self, expected_containers,
                                  group=None,
                                  full_name=True, bool_check=False,
                                  nodenames=None):
        group = group or ALL_COMPUTES_GROUP_NAME
        return containers.assert_containers_running(
            group=group,
            expected_containers=expected_containers,
            full_name=full_name,
            bool_check=bool_check,
            nodenames=nodenames)

    def list_containers_df(self, group=None):
        return containers.list_containers_df(group)

    def add_node(self,
                 hostname: typing.Optional[str] = None,
                 address: typing.Optional[str] = None,
                 group: typing.Optional[str] = None,
                 ssh_client: typing.Optional[ssh.SSHClientFixture] = None,
                 **create_params) \
            -> topology.OpenStackTopologyNode:
        node = super(PodifiedTopology, self).add_node(
            hostname=hostname,
            address=address,
            group=group,
            ssh_client=ssh_client,
            **create_params
        )
        # NOTE(slaweq): additionally lets add every edpm node to the "legacy"
        # group named "compute"
        if group and group in COMPUTE_GROUPS:
            group_nodes = self.add_group(group=ALL_COMPUTES_GROUP_NAME)
            if node and node not in group_nodes:
                group_nodes.append(node)
                node.add_group(group=ALL_COMPUTES_GROUP_NAME)
        return node

    def create_node(self, name, ssh_client, **kwargs):
        node_type = kwargs.pop('node_type')
        if node_type == OCP_WORKER:
            return OcpWorkerNode(topology=self, name=name, ssh_client=None,
                                 **kwargs)
        else:
            return EdpmNode(topology=self, name=name, ssh_client=ssh_client,
                            **kwargs)

    def discover_nodes(self):
        self.discover_ssh_proxy_jump_node()
        self.discover_ocp_worker_nodes()
        self.discover_edpm_nodes()

    def discover_ssh_proxy_jump_node(self):
        pass

    def discover_ocp_worker_nodes(self):
        # NOTE(slaweq): For now this will only discover nodes but there will be
        # no ssh_client created to ssh to those nodes. Getting
        # ssh_client to those nodes may be implemented in the future if that
        # will be needed, but this may be hard e.g. for the CRC environments as
        # in that case internal OCP worker's IP address is not accessible from
        # outside at all
        for worker_data in _openshift.list_ocp_workers():
            node = self._add_node(
                addresses=worker_data['addresses'],
                hostname=worker_data['hostname'],
                ssh_client=None,
                create_ssh_client=False,
                node_type=OCP_WORKER)
            group_nodes = self.add_group(group='controller')
            if node not in group_nodes:
                group_nodes.append(node)
                node.add_group(group='controller')

    def discover_edpm_nodes(self):
        for node in _openshift.list_edpm_nodes():
            LOG.debug(f"Found EDPM node {node['hostname']} "
                      f"(IP: {node['host']})")
            group = node.pop('group')
            host_config = _edpm.edpm_host_config(node)
            ssh_client = _edpm.edpm_ssh_client(host_config=host_config)
            node = self.add_node(address=host_config.host,
                                 group=group,
                                 ssh_client=ssh_client,
                                 node_type=EDPM_NODE)
            assert isinstance(node, EdpmNode)


class EdpmNode(rhosp.RhospNode):

    def power_on_node(self):
        LOG.debug(f"Ensuring EDPM node {self.name} power is on...")
        self.ssh_client.close()
        _openshift.power_on_edpm_node(self.name)
        hostname = sh.get_hostname(ssh_client=self.ssh_client)
        LOG.debug(f"Overcloud node {self.name} power is on ("
                  f"hostname={hostname})")

    def power_off_node(self):
        LOG.debug(f"Ensuring EDPM node {self.name} power is off...")
        self.ssh_client.close()
        _openshift.power_off_edpm_node(self.name)
        LOG.debug(f"EDPM node {self.name} power is off.")


class OcpWorkerNode(rhosp.RhospNode):
    pass


def setup_podified_topology():
    topology.set_default_openstack_topology_class(PodifiedTopology)



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import


from tobiko.shell import sh
from tobiko.tests import unit


SPECIAL_CHARS = r'@&%+=:,.;<>/-()[]*|~'


class ShellCommandTest(unit.TobikoUnitTest):

    def test_from_str(self):
        result = sh.shell_command('ls -lh *.py')
        self.assertIsInstance(result, sh.ShellCommand)
        self.assertEqual(('ls', '-lh', '*.py'), result)
        self.assertEqual("ls -lh *.py", str(result))

    def test_from_str_with_quotes(self):
        result = sh.shell_command('ls -lh "with quotes"')
        self.assertIsInstance(result, sh.ShellCommand)
        self.assertEqual(('ls', '-lh', "with quotes"), result)
        self.assertEqual("ls -lh 'with quotes'", str(result))

    def test_from_sequence(self):
        result = sh.shell_command(['ls', '-lh', '*.py'])
        self.assertIsInstance(result, sh.ShellCommand)
        self.assertEqual(('ls', '-lh', '*.py'), result)
        self.assertEqual("ls -lh *.py", str(result))

    def test_from_sequence_with_quotes(self):
        result = sh.shell_command(['ls', '-lh', "with quotes"])
        self.assertIsInstance(result, sh.ShellCommand)
        self.assertEqual(('ls', '-lh', "with quotes"), result)
        self.assertEqual("ls -lh 'with quotes'", str(result))

    def test_from_shell_command(self):
        other = sh.shell_command(['ls', '-lh', '*.py'])
        result = sh.shell_command(other)
        self.assertIs(other, result)

    def test_from_special_chars(self):
        command = sh.shell_command(SPECIAL_CHARS)
        self.assertEqual((SPECIAL_CHARS,), command)
        self.assertEqual(SPECIAL_CHARS, str(command))

    def test_from_journalctl_command(self):
        command_line = (
            'journalctl', '--no-pager', '--unit',
            'devstack@q-svc', '--since', '30 minutes ago',
            '--output', 'short-iso', '--grep',
            "'Nova.+event.+response.*09e69236-2a3b-4077-bd50-0c80946bf5b3'")
        command = sh.shell_command(command_line)
        self.assertEqual(command_line, command)
        self.assertEqual(
            "journalctl --no-pager --unit devstack@q-svc "
            "--since '30 minutes ago' --output short-iso --grep "
            "'Nova.+event.+response.*09e69236-2a3b-4077-bd50-0c80946bf5b3'",
            str(command))

    def test_add_str(self):
        base = sh.shell_command('ssh pippo@clubhouse.mouse')
        result = base + 'ls -lh *.py'
        self.assertEqual(('ssh', 'pippo@clubhouse.mouse', 'ls', '-lh', '*.py'),
                         result)
        self.assertEqual("ssh pippo@clubhouse.mouse ls -lh *.py",
                         str(result))

    def test_add_str_with_quotes(self):
        base = sh.shell_command('sh -c')
        result = base + "'echo Hello!'"
        self.assertIsInstance(result, sh.ShellCommand)
        self.assertEqual(('sh', '-c', "echo Hello!"), result)
        self.assertEqual("sh -c 'echo Hello!'", str(result))

    def test_add_sequence(self):
        base = sh.shell_command('ssh pippo@clubhouse.mouse')
        result = base + ['ls', '-lh', '*.py']
        self.assertEqual(('ssh', 'pippo@clubhouse.mouse', 'ls', '-lh', '*.py'),
                         result)
        self.assertEqual("ssh pippo@clubhouse.mouse ls -lh *.py",
                         str(result))

    def test_add_sequence_with_quotes(self):
        base = sh.shell_command('sh -c')
        result = base + ['echo Hello!']
        self.assertIsInstance(result, sh.ShellCommand)
        self.assertEqual(('sh', '-c', "echo Hello!"), result)
        self.assertEqual("sh -c 'echo Hello!'", str(result))

    def test_add_shell_command(self):
        base = sh.shell_command('ssh pippo@clubhouse.mouse')
        result = base + sh.shell_command(['ls', '-lh', '*.py'])
        self.assertEqual(('ssh', 'pippo@clubhouse.mouse', 'ls', '-lh', '*.py'),
                         result)
        self.assertEqual("ssh pippo@clubhouse.mouse ls -lh *.py",
                         str(result))

    def test_add_shell_command_with_quotes(self):
        base = sh.shell_command('sh -c')
        result = base + sh.shell_command(['echo Hello!'])
        self.assertIsInstance(result, sh.ShellCommand)
        self.assertEqual(('sh', '-c', "echo Hello!"), result)
        self.assertEqual("sh -c 'echo Hello!'", str(result))

    def test_add_special_chars(self):
        base = sh.shell_command('echo')
        result = base + SPECIAL_CHARS
        self.assertEqual(('echo', SPECIAL_CHARS), result)
        self.assertEqual('echo ' + SPECIAL_CHARS, str(result))



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import io
import sys
import typing
import unittest

import tobiko


def get_test_result() -> unittest.TestResult:
    return tobiko.setup_fixture(TestResultFixture).result


class TestResultFixture(tobiko.SharedFixture):

    verbosity = 2
    stream = sys.stderr
    description = True
    result: unittest.TestResult

    def setup_fixture(self):
        self.result = TestResult(stream=self.stream,
                                 verbosity=self.verbosity,
                                 description=self.description)


class TestResult(unittest.TextTestResult):

    def __init__(self,
                 stream: typing.TextIO,
                 description: bool,
                 verbosity: int):
        super().__init__(stream=TextIOWrapper(stream),
                         descriptions=description,
                         verbosity=verbosity)
        self.buffer = True

    def startTest(self, test: unittest.TestCase):
        tobiko.push_test_case(test)
        super().startTest(test)

    def stopTest(self, test: unittest.TestCase) -> None:
        super().stopTestRun()
        actual_test = tobiko.pop_test_case()
        assert actual_test == test
        tobiko.remove_test_from_all_shared_resources(test.id())


class TextIOWrapper(io.TextIOWrapper):

    def __init__(self, stream: typing.TextIO):
        super().__init__(buffer=stream.buffer,
                         encoding='UTF-8',
                         errors='strict',
                         line_buffering=True,
                         write_through=False)

    def writeln(self, line: str):
        self.write(line + '\n')



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.octavia import _amphora
from tobiko.openstack.octavia import _deployers
from tobiko.openstack.octavia import _client
from tobiko.openstack.octavia import _constants
from tobiko.openstack.octavia import _exceptions
from tobiko.openstack.octavia import _validators
from tobiko.openstack.octavia import _waiters

AmphoraIdType = _amphora.AmphoraIdType
AmphoraType = _amphora.AmphoraType
get_amphora_compute_node = _amphora.get_amphora_compute_node
get_master_amphora = _amphora.get_master_amphora
list_amphorae = _amphora.list_amphorae
get_amphora_stats = _amphora.get_amphora_stats
get_amphora = _amphora.get_amphora
run_command_on_amphora = _amphora.run_command_on_amphora

OCTAVIA_CLIENT_CLASSSES = _client.OCTAVIA_CLIENT_CLASSSES
get_octavia_client = _client.get_octavia_client
octavia_client = _client.octavia_client
OctaviaClientFixture = _client.OctaviaClientFixture
OctaviaClientType = _client.OctaviaClientType
list_members = _client.list_members
list_load_balancers = _client.list_load_balancers
find_load_balancer = _client.find_load_balancer
create_load_balancer = _client.create_load_balancer
find_listener = _client.find_listener
create_listener = _client.create_listener
find_pool = _client.find_pool
create_pool = _client.create_pool
find_member = _client.find_member
create_member = _client.create_member

# Waiters
wait_for_status = _waiters.wait_for_status
wait_for_octavia_service = _waiters.wait_for_octavia_service

# Validators
check_members_balanced = _validators.check_members_balanced

# Exceptions
RequestException = _exceptions.RequestException
TimeoutException = _exceptions.TimeoutException
OctaviaClientException = _exceptions.OctaviaClientException
RoundRobinException = _exceptions.RoundRobinException
TrafficTimeoutError = _exceptions.TrafficTimeoutError
AmphoraMgmtPortNotFound = _exceptions.AmphoraMgmtPortNotFound

# Constants
PROVISIONING_STATUS = _constants.PROVISIONING_STATUS
OPERATING_STATUS = _constants.OPERATING_STATUS
ACTIVE = _constants.ACTIVE
ERROR = _constants.ERROR
PENDING_UPDATE = _constants.PENDING_UPDATE
ONLINE = _constants.ONLINE
PROTOCOL_HTTP = _constants.PROTOCOL_HTTP
PROTOCOL_TCP = _constants.PROTOCOL_TCP
LB_ALGORITHM_ROUND_ROBIN = _constants.LB_ALGORITHM_ROUND_ROBIN
LB_ALGORITHM_SOURCE_IP_PORT = _constants.LB_ALGORITHM_SOURCE_IP_PORT
AMPHORA_PROVIDER = _constants.AMPHORA_PROVIDER
OVN_PROVIDER = _constants.OVN_PROVIDER
OCTAVIA_PROVIDERS_NAMES = _constants.OCTAVIA_PROVIDERS_NAMES
WORKER_SERVICE = _constants.WORKER_SERVICE
HOUSEKEEPING_SERVICE = _constants.HOUSEKEEPING_SERVICE
HM_SERVICE = _constants.HM_SERVICE
API_SERVICE = _constants.API_SERVICE
WORKER_CONTAINER = _constants.WORKER_CONTAINER
HOUSEKEEPING_CONTAINER = _constants.HOUSEKEEPING_CONTAINER
HM_CONTAINER = _constants.HM_CONTAINER
API_CONTAINER = _constants.API_CONTAINER
OCTAVIA_SERVICES = [WORKER_SERVICE, HOUSEKEEPING_SERVICE, HM_SERVICE,
                    API_SERVICE]
OCTAVIA_CONTAINERS = [WORKER_CONTAINER, HOUSEKEEPING_CONTAINER, HM_CONTAINER,
                      API_CONTAINER]
LB_AMP_NAME = _constants.LB_AMP_NAME
LISTENER_AMP_NAME = _constants.LISTENER_AMP_NAME
POOL_AMP_NAME = _constants.POOL_AMP_NAME
MEMBER_AMP_NAME_PREFIX = _constants.MEMBER_AMP_NAME_PREFIX
LB_OVN_NAME = _constants.LB_OVN_NAME
LISTENER_OVN_NAME = _constants.LISTENER_OVN_NAME
POOL_OVN_NAME = _constants.POOL_OVN_NAME
MEMBER_OVN_NAME_PREFIX = _constants.MEMBER_OVN_NAME_PREFIX

# Deployers
deploy_ipv4_amphora_lb = _deployers.deploy_ipv4_amphora_lb
deploy_ipv4_ovn_lb = _deployers.deploy_ipv4_ovn_lb




# Copyright 2023 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import base64
import io
import os

import tobiko
from tobiko import config
from tobiko.podified import _openshift
from tobiko.shell import ssh

CONF = config.CONF


def edpm_host_config(node=None,
                     ip_version: int = None,
                     key_filename: str = None):
    node = node or {}
    host_config = EdpmHostConfig(
        host=node.get('host'),
        ip_version=ip_version,
        key_filename=key_filename,
        port=node.get('port'),
        username=node.get('username'))
    return tobiko.setup_fixture(host_config)


def edpm_ssh_client(ip_version: int = None,
                    host_config=None,
                    node=None):
    if host_config is None:
        host_config = edpm_host_config(node=node,
                                       ip_version=ip_version)
    tobiko.check_valid_type(host_config.host, str)
    return ssh.ssh_client(host=host_config.host,
                          **host_config.connect_parameters)


class EdpmSshKeyFileFixture(tobiko.SharedFixture):

    @property
    def key_filename(self):
        return tobiko.tobiko_config_path(
            CONF.tobiko.rhosp.ssh_key_filename)

    def setup_fixture(self):
        self.setup_key_file()

    def setup_key_file(self):
        priv_key_filename = self.key_filename
        pub_key_filename = priv_key_filename + ".pub"
        key_dirname = os.path.dirname(priv_key_filename)
        tobiko.makedirs(key_dirname, mode=0o700)

        private_key, public_key = _openshift.get_dataplane_ssh_keypair()
        if private_key:
            with io.open(priv_key_filename, 'wb') as fd:
                fd.write(base64.b64decode(private_key))
                os.chmod(priv_key_filename, 0o600)
        if public_key:
            with io.open(pub_key_filename, 'wb') as fd:
                fd.write(base64.b64decode(public_key))
                os.chmod(pub_key_filename, 0o600)


class EdpmHostConfig(tobiko.SharedFixture):

    key_file = tobiko.required_fixture(EdpmSshKeyFileFixture)

    def __init__(self,
                 host: str,
                 ip_version: int = None,
                 key_filename: str = None,
                 port: int = None,
                 username: str = None,
                 **kwargs):
        super(EdpmHostConfig, self).__init__()
        self.host = host
        self.ip_version = ip_version
        self.key_filename = key_filename
        self.port = port
        self.username = username
        self._connect_parameters = ssh.gather_ssh_connect_parameters(**kwargs)

    def setup_fixture(self):
        if self.port is None:
            self.port = CONF.tobiko.rhosp.ssh_port
        if self.username is None:
            self.username = CONF.tobiko.rhosp.ssh_username
        if self.key_filename is None:
            self.key_filename = self.key_file.key_filename

    @property
    def connect_parameters(self):
        parameters = ssh.ssh_host_config(
            host=str(self.host)).connect_parameters
        parameters.update(ssh.gather_ssh_connect_parameters(self))
        parameters.update(self._connect_parameters)
        return parameters



from __future__ import absolute_import

import time
import typing  # noqa
from functools import wraps


from oslo_log import log
import pandas

import tobiko
from tobiko import tripleo
from tobiko.tripleo import overcloud
from tobiko.shell import ping
from tobiko.shell import sh
from tobiko.openstack import nova
from tobiko.openstack import topology
from tobiko.openstack import stacks
from tobiko.tripleo import containers


LOG = log.getLogger(__name__)


def check_nova_services_health(timeout=600., interval=2.):
    retry = tobiko.retry(timeout=timeout, interval=interval)
    nova.wait_for_services_up(retry=retry)


def start_all_instances():
    """try to start all stopped overcloud instances"""
    for instance in nova.list_servers():
        activated_instance = nova.activate_server(instance)
        time.sleep(3)
        instance_info = 'instance {nova_instance} is {state} on {host}'.format(
            nova_instance=activated_instance.name,
            state=activated_instance.status,
            host=activated_instance._info[  # pylint: disable=W0212
                'OS-EXT-SRV-ATTR:hypervisor_hostname'])
        LOG.info(instance_info)
        if activated_instance.status != 'ACTIVE':
            tobiko.fail(instance_info)


def stop_all_instances():
    """try to start all stopped overcloud instances"""
    for instance in nova.list_servers():
        activated_instance = nova.shutoff_server(instance)
        time.sleep(3)
        instance_info = 'instance {nova_instance} is {state} on {host}'.format(
            nova_instance=activated_instance.name,
            state=activated_instance.status,
            host=activated_instance._info[  # pylint: disable=W0212
                'OS-EXT-SRV-ATTR:hypervisor_hostname'])
        LOG.info(instance_info)
        if activated_instance.status != 'SHUTOFF':
            tobiko.fail(instance_info)


def wait_for_all_instances_status(status, timeout=None):
    """wait for all instances for a certain status or raise an exception"""
    for instance in nova.list_servers():
        nova.wait_for_server_status(server=instance.id, status=status,
                                    timeout=timeout)
        instance_info = 'instance {nova_instance} is {state} on {host}'.format(
            nova_instance=instance.name,
            state=status,
            host=instance._info[  # pylint: disable=W0212
                'OS-EXT-SRV-ATTR:hypervisor_hostname'])
        LOG.info(instance_info)


def get_vms_table():
    """populate a dataframe with vm host,id,status"""
    vms_data = [(vm._info[  # pylint: disable=W0212
                     'OS-EXT-SRV-ATTR:hypervisor_hostname'], vm.id,
                 vm.status) for vm in nova.list_servers()]
    vms_df = pandas.DataFrame(vms_data, columns=['vm_host', 'vm_id',
                                                 'vm_state'])
    return vms_df


def list_computes():
    """list compute host names"""
    return [compute.hypervisor_hostname for compute in nova.list_hypervisors()]


def get_compute_vms_df(compute_host):
    """input: compute hostname (can be short)
    output: dataframe with vms of that host"""
    return get_vms_table().query(f"vm_host=='{compute_host}'")


def get_random_compute_with_vms_name():
    """get a randomcompute holding vm/s"""
    for compute in list_computes():
        if not get_compute_vms_df(compute).empty:
            return compute


def vm_info(vm_id, vms_df):
    """input: vm and a vms df
    output: host string"""
    return vms_df.query(f"vm_id == '{vm_id}'").to_string()


def vm_df(vm_id, vms_df):
    """input: vm and a vms df
    output: host string"""
    return vms_df.query(f"vm_id == '{vm_id}'")


def vm_floating_ip(vm_id):
    """input: vm_id
    output it's floating ip"""

    vm = nova.get_server(vm_id)
    floating_ip = nova.list_server_ip_addresses(
        vm, address_type='floating').first
    return floating_ip


def check_ping_vm_fip(fip):
    ping.ping_until_received(fip).assert_replied()


def check_df_vms_ping(df):
    """input: dataframe with vms_ids
    try to ping all vms in df"""

    for vm_id in df.vm_id.to_list():
        check_ping_vm_fip(vm_floating_ip(vm_id))


def vm_location(vm_id, vms_df):
    """input: vm and a vms df
    output: host string"""
    return vms_df.query(f"vm_id == '{vm_id}'")['vm_host'].to_string(
            index=False)


def check_vm_evacuations(vms_df_old=None, compute_host=None, timeout=600,
                         interval=2, check_no_evacuation=False):
    """check evacuation of vms
    input: old and new vms_state_tables dfs"""
    failures = []
    start = time.time()

    while time.time() - start < timeout:
        failures = []
        vms_df_new = get_compute_vms_df(compute_host)
        for vm_id in vms_df_old.vm_id.to_list():
            old_bm_host = vm_location(vm_id, vms_df_old)
            new_vm_host = vm_location(vm_id, vms_df_new)

            if check_no_evacuation:
                cond = bool(old_bm_host != new_vm_host)
            else:
                cond = bool(old_bm_host == new_vm_host)

            if cond:
                failures.append(
                    'failed vm evacuations: {}\n\n'.format(vm_info(vm_id,
                                                           vms_df_old)))
            if failures:
                LOG.info('Failed nova evacuation:\n {}'.format(failures))
                LOG.info('Not all nova vms evacuated ..')
                LOG.info('Retrying , timeout at: {}'
                         .format(timeout-(time.time() - start)))
                time.sleep(interval)
            else:
                LOG.info(vms_df_old.to_string())
                LOG.info('All vms were evacuated!')
                return
    # exhausted all retries
    if failures:
        tobiko.fail(
            'failed vm evacuations:\n{!s}', '\n'.join(failures))


def get_stack_server_id(stack):
    return stack.server_details.id


def get_fqdn_from_topology_node(topology_node):
    return sh.execute("hostname -f", ssh_client=topology_node.ssh_client,
                      expect_exit_status=None).stdout.strip()


def check_vm_running_via_virsh(topology_compute, vm_id):
    """check that a vm is in running state via virsh command,
    return false if not"""
    if vm_id in get_vm_uuid_list_running_via_virsh(topology_compute):
        return True
    else:
        return False


def get_vm_uuid_list_running_via_virsh(topology_compute):
    if overcloud.has_overcloud():
        container_runtime = containers.get_container_runtime_name()
        nova_libvirt = containers.get_libvirt_container_name()
        command = f"sudo {container_runtime} exec {nova_libvirt} " \
                  f"sh -c 'for i in `virsh list --name --state-running` " \
                  f";do virsh domuuid $i;done'"
    else:
        command = "for i in `sudo virsh list --name --state-running` " \
                  ";do virsh domuuid $i;done'"
    return sh.execute(command,
                      ssh_client=topology_compute.ssh_client).stdout.split()


def check_computes_vms_running_via_virsh():
    """check all vms are running via virsh list command"""
    for compute in topology.list_openstack_nodes(group='compute'):
        hostname = get_fqdn_from_topology_node(compute)
        retry = tobiko.retry(timeout=120, interval=5)
        vms_df = get_compute_vms_df(hostname)
        for vm_id in vms_df.vm_id.to_list():
            for _ in retry:
                if check_vm_running_via_virsh(compute, vm_id):
                    LOG.info(f"{vm_id} is running ok on "
                             f"{compute.hostname}")
                    break
                else:
                    LOG.info(f"{vm_id} is not in running state on "
                             f"{compute.hostname}")


def get_nova_server_floating_ip():
    """get an a running's vm floating_ip"""
    return tobiko.setup_fixture(
           stacks.CirrosServerStackFixture).floating_ip_address


# Test is inteded for D/S env
@tripleo.skip_if_missing_overcloud
def check_or_start_background_vm_ping():
    """Check if process exists, if so stop and check ping health
    if not : start a new separate ping process.
    Executes a Background ping to a vm floating_ip,
    this test is intended to be run and picked up again
    by the next tobiko run. Ping results are parsed
    and a failure is raised if ping failure is above a certain amount"""
    ping_vm_fip = get_nova_server_floating_ip()
    sh.check_or_start_background_process(
        bg_function=ping.write_ping_to_file,
        bg_process_name='tobiko_background_ping',
        check_function=ping.check_ping_statistics,
        ping_ip=ping_vm_fip)


# Test is inteded for D/S env
@tripleo.skip_if_missing_overcloud
def skip_check_or_start_background_vm_ping():
    """Like the above, but skips the ping check, truncates results
    and reexecutes the test"""
    ping_vm_fip = get_nova_server_floating_ip()
    sh.check_or_start_background_process(
        bg_function=ping.write_ping_to_file,
        bg_process_name='tobiko_background_ping',
        check_function=ping.skip_check_ping_statistics,
        ping_ip=ping_vm_fip)


def skip_background_vm_ping_checks(func):
    """Skip ping_check_decorator - to be used when traffic to vm
    must be dropped for the duration of the test - func"""
    @wraps(func)
    def wrapper(*args):  # pylint: disable=W0613
        tobiko.add_cleanup(skip_check_or_start_background_vm_ping)
        check_or_start_background_vm_ping()
        func(*args)
    return wrapper


def skip_background_vm_ping_checks_when_nondvr(func):
    """Similar to skip_background_vm_ping_checks, but the background ping
    checks and the restart of the background ping process is only executed when
    DVR is disabled"""
    @wraps(func)
    def wrapper(*args):  # pylint: disable=W0613
        if not overcloud.is_dvr_enabled():
            tobiko.add_cleanup(skip_check_or_start_background_vm_ping)
            check_or_start_background_vm_ping()
        func(*args)
    return wrapper



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import netaddr
import testtools

from tobiko import shiftstack
from tobiko.shell import ping
from tobiko.shiftstack import stacks


@shiftstack.skip_unless_has_shiftstack()
class ShiftstackStacksTest(testtools.TestCase):

    def test_ensure_shiftstack_node_floating_ip(self):
        nodes = shiftstack.list_shiftstack_nodes(status='ACTIVE')
        floating_ips = []
        for node in nodes:
            floating_ip = stacks.ensure_shiftstack_node_floating_ip(
                server=node)
            self.assertIsInstance(floating_ip, netaddr.IPAddress)
            floating_ips.append(floating_ip)
        ping.assert_reachable_hosts(floating_ips)



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import random

import testtools
from oslo_log import log

import tobiko
from tobiko.openstack import neutron
from tobiko.openstack import openstackclient
from tobiko.openstack import stacks


LOG = log.getLogger(__name__)


class BaseCliTest(testtools.TestCase):

    def setUp(self):
        super(BaseCliTest, self).setUp()
        self.api = neutron.get_neutron_client()

    def api_network_delete(self, network):
        nets = self.api.list_networks()['networks']
        for net in nets:
            if net['name'] == network:
                self.api.delete_network(net['id'])
                break
            if net['id'] == network:
                self.api.delete_network(network)
                break

    def api_subnet_delete(self, subnet_name):
        subnets = self.api.list_subnets()['subnets']
        for subnet in subnets:
            if subnet['name'] == subnet_name:
                self.api.delete_subnet(subnet['id'])
                break
            if subnet['id'] == subnet_name:
                self.api.delete_subnet(subnet_name)
                break

    def api_port_delete(self, port_name):
        ports = self.api.list_ports()['ports']
        for port in ports:
            if port['name'] == port_name:
                self.api.delete_port(port['id'])
                break
            if port['id'] == port_name:
                self.api.delete_port(port_name)
                break

    def api_network_log_delete(self, log_name):
        logs = self.api.list_network_logs()['logs']
        for _log in logs:
            if _log['name'] == log_name:
                self.api.delete_network_log(_log['id'])
                break
            if _log['id'] == log_name:
                self.api.delete_network_log(_log['id'])
                break

    def api_random_port_create(self):
        net_name = self.random_name()
        port_name = self.random_name()
        network = self.api.create_network({'network': {'name': net_name}})
        self.addCleanup(self.api_network_delete, net_name)
        network_id = network['network']['id']
        self.api.create_port({'port': {'name': port_name,
                                       'network_id': network_id}})
        self.addCleanup(self.api_port_delete, port_name)
        return port_name

    def api_random_subnet_create(self):
        net_name = self.random_name()
        subnet_name = self.random_name()
        network = self.api.create_network({'network': {'name': net_name}})
        self.addCleanup(self.api_network_delete, net_name)
        network_id = network['network']['id']
        self.api.create_subnet({'subnet': {'name': subnet_name,
                                           'network_id': network_id,
                                           'ip_version': 4,
                                           'cidr': '123.123.123.0/24'}})
        return subnet_name

    def api_random_network_create(self):
        name = self.random_name()
        self.api.create_network({'network': {'name': name}})
        self.addCleanup(self.api_network_delete, name)
        return name

    def api_random_network_log_create(self, sec_group_id=None):
        name = self.random_name()
        api_args = {'log': {'name': name,
                    'resource_type': 'security_group'}}
        if sec_group_id:
            api_args['log']['resource_id'] = sec_group_id
        self.api.create_network_log(api_args)
        self.addCleanup(self.api_network_log_delete, name)
        return name

    def random_name(self, length=16):
        letters = 'abcdefghijklmnopqrstuvwxyz'
        random_string = ''.join(random.choice(letters) for i in range(length))
        return f'{self.__class__.__name__}-{random_string}'


@neutron.skip_if_missing_networking_extensions('logging')
class NeutronLogCliTest(BaseCliTest):

    LOGS_AMOUNT = 2

    sec_groups_stack = tobiko.required_fixture(
        stacks.SecurityGroupsFixture)

    def _get_icmp_sec_group_id(self):
        """Returns the uuid of a security group with ICMP allowed."""
        sec_group_obj = self.sec_groups_stack.icmp_security_group_id
        if hasattr(sec_group_obj, '__iter__'):
            return sec_group_obj
        return sec_group_obj[0]

    def test_network_loggable_resources_list(self):
        response = openstackclient.network_loggable_resources_list()
        self.assertIn('security_group', response[0]['Supported types'],
                      "Security group logging isn't supported.")

    def test_network_log_create(self):
        log_name = self.random_name()
        test_sec_group_id = self._get_icmp_sec_group_id()
        response = openstackclient.network_log_create(
                log_name, **{'resource-type': 'security_group',
                             'resource': test_sec_group_id,
                             'event': 'ALL'})
        self.addCleanup(self.api_network_log_delete, log_name)
        err_msg = 'Creation of log for security group using CLI failed.'
        # pylint: disable=E1126
        self.assertEqual(response['Enabled'], True, err_msg)
        self.assertEqual(response['Type'], 'security_group', err_msg)
        self.assertEqual(response['Event'], 'ALL', err_msg)

    def test_network_log_show(self):
        test_sec_group_id = self._get_icmp_sec_group_id()
        log_name = self.api_random_network_log_create(test_sec_group_id)
        _log = openstackclient.network_log_show(log_name)
        err_msg = 'Details show of log for security group using CLI failed.'
        # pylint: disable=E1126
        self.assertEqual(_log['Name'], log_name, err_msg)

    def test_network_log_create_all(self):
        log_name = self.random_name()
        response = openstackclient.network_log_create(
                log_name, **{'resource-type': 'security_group',
                             'event': 'ALL'})
        self.addCleanup(self.api_network_log_delete, log_name)
        err_msg = 'Creation of log for security group using CLI failed.'
        # pylint: disable=E1126
        self.assertEqual(response['Enabled'], True, err_msg)
        self.assertEqual(response['Type'], 'security_group', err_msg)
        self.assertEqual(response['Event'], 'ALL', err_msg)

    def test_network_log_list(self):
        test_sec_group_id = self._get_icmp_sec_group_id()
        log_names = [self.api_random_network_log_create(test_sec_group_id)
                     for i in range(self.LOGS_AMOUNT)]
        for log_name in log_names:
            self.addCleanup(self.api_network_log_delete, log_name)
        logs = openstackclient.network_log_list()
        found_count = 0
        for _log in logs:
            if _log['Name'] in log_names:
                found_count += 1
        err_msg = (f'Listing {self.LOGS_AMOUNT} logs for security groups '
                   'using CLI failed.')
        # pylint: disable=E1126
        self.assertEqual(found_count, self.LOGS_AMOUNT, err_msg)

    def test_network_log_delete(self):
        test_sec_group_id = self._get_icmp_sec_group_id()
        log_names = [self.api_random_network_log_create(test_sec_group_id)
                     for i in range(self.LOGS_AMOUNT)]
        openstackclient.network_log_delete(log_names)
        logs = self.api.list_network_logs()['logs']
        err_msg = 'Deletion of logs for security group using CLI failed.'
        for log_name in log_names:
            for _log in logs:
                self.assertNotEqual(_log['name'], log_name, err_msg)


class NeutronNetworkCliTest(BaseCliTest):

    def test_network_creation(self):
        net_name = self.random_name()
        output = openstackclient.network_create(net_name)
        self.addCleanup(self.api_network_delete, net_name)
        self.assertEqual(output['name'], net_name)  # pylint: disable=E1126
        self.assertEqual(output['status'], 'ACTIVE')  # pylint: disable=E1126

    def test_network_deletion(self):
        net_name_1 = self.api_random_network_create()
        net_name_2 = self.api_random_network_create()
        openstackclient.network_delete([net_name_1, net_name_2])
        nets = self.api.list_networks()['networks']
        for net in nets:
            self.assertNotEqual(net['name'], net_name_1)
            self.assertNotEqual(net['name'], net_name_2)

    def test_network_list(self):
        net_name = self.api_random_network_create()
        nets = openstackclient.network_list()
        found = False
        for net in nets:
            if net['Name'] == net_name:
                found = True
                break
        self.assertTrue(found)

    def test_network_show(self):
        net_name = self.api_random_network_create()
        net = openstackclient.network_show(net_name)
        self.assertEqual(net['name'], net_name)  # pylint: disable=E1126


class NeutronSubnetCliTest(BaseCliTest):

    def test_subnet_creation(self):
        subnet_name = self.random_name()
        net_name = self.api_random_network_create()
        output = openstackclient.subnet_create(
                subnet_name, net_name, **{'subnet-range': '123.123.123.0/24'})
        self.assertEqual(output['name'], subnet_name)  # pylint: disable=E1126

    def test_subnet_deletion(self):
        subnet_name_1 = self.api_random_subnet_create()
        subnet_name_2 = self.api_random_subnet_create()
        openstackclient.subnet_delete([subnet_name_1, subnet_name_2])
        subnets = self.api.list_subnets()['subnets']
        for subnet in subnets:
            self.assertNotEqual(subnet['name'], subnet_name_1)
            self.assertNotEqual(subnet['name'], subnet_name_2)

    def test_subnet_list(self):
        subnet_name = self.api_random_subnet_create()
        subnets = openstackclient.subnet_list()
        found = False
        for subnet in subnets:
            if subnet['Name'] == subnet_name:
                found = True
                break
        self.assertTrue(found)

    def test_subnet_show(self):
        subnet_name = self.api_random_subnet_create()
        subnet = openstackclient.subnet_show(subnet_name)
        self.assertEqual(subnet['name'], subnet_name)  # pylint: disable=E1126


class NeutronPortCliTest(BaseCliTest):

    def test_port_creation(self):
        port_name = self.random_name()
        net_name = self.api_random_network_create()
        output = openstackclient.port_create(port_name, net_name)
        self.addCleanup(self.api_port_delete, port_name)
        self.assertEqual(output['name'], port_name)  # pylint: disable=E1126

    def test_port_deletion(self):
        port_name_1 = self.api_random_port_create()
        port_name_2 = self.api_random_port_create()
        openstackclient.port_delete([port_name_1, port_name_2])
        ports = self.api.list_ports()['ports']
        for port in ports:
            self.assertNotEqual(port['name'], port_name_1)
            self.assertNotEqual(port['name'], port_name_2)

    def test_port_list(self):
        port_name = self.api_random_port_create()
        ports = openstackclient.port_list()
        found = False
        for port in ports:
            if port['Name'] == port_name:
                found = True
                break
        self.assertTrue(found)

    def test_port_show(self):
        port_name = self.api_random_port_create()
        port = openstackclient.port_show(port_name)
        self.assertEqual(port['name'], port_name)  # pylint: disable=E1126



"""Exported method Container.attach()."""

from __future__ import absolute_import

import collections
import fcntl
import logging
import struct
import sys
import termios


class Mixin:
    """Publish attach() for inclusion in Container class."""

    def attach(self, eot=4, stdin=None, stdout=None):
        """Attach to container's PID1 stdin and stdout.

        stderr is ignored.
        PseudoTTY work is done in start().
        """
        if stdin is None:
            stdin = sys.stdin.fileno()
        elif hasattr(stdin, 'fileno'):
            stdin = stdin.fileno()

        if stdout is None:
            stdout = sys.stdout.fileno()
        elif hasattr(stdout, 'fileno'):
            stdout = stdout.fileno()

        with self._client() as podman:
            attach = podman.GetAttachSockets(self._id)

        # This is the UDS where all the IO goes
        io_socket = attach['sockets']['io_socket']
        assert len(io_socket) <= 107,\
            'Path length for sockets too long. {} > 107'.format(
                len(io_socket)
            )

        # This is the control socket where resizing events are sent to conmon
        # attach['sockets']['control_socket']
        self.pseudo_tty = collections.namedtuple(
            'PseudoTTY',
            ['stdin', 'stdout', 'io_socket', 'control_socket', 'eot'])(
                stdin,
                stdout,
                attach['sockets']['io_socket'],
                attach['sockets']['control_socket'],
                eot,
            )

    @property
    def resize_handler(self):
        """Send the new window size to conmon."""

        def wrapped(signum, frame):  # pylint: disable=unused-argument
            packed = fcntl.ioctl(self.pseudo_tty.stdout, termios.TIOCGWINSZ,
                                 struct.pack('HHHH', 0, 0, 0, 0))
            rows, cols, _, _ = struct.unpack('HHHH', packed)
            logging.debug('Resize window(%dx%d) using %s', rows, cols,
                          self.pseudo_tty.control_socket)

            # TODO: Need some kind of timeout in case pipe is blocked
            with open(self.pseudo_tty.control_socket, 'w') as skt:
                # send conmon window resize message
                skt.write('1 {} {}\n'.format(rows, cols))

        return wrapped

    @property
    def log_handler(self):
        """Send command to reopen log to conmon."""

        def wrapped(signum, frame):  # pylint: disable=unused-argument
            with open(self.pseudo_tty.control_socket, 'w') as skt:
                # send conmon reopen log message
                skt.write('2\n')

        return wrapped



from __future__ import absolute_import

import testtools
from oslo_log import log

import tobiko
from tobiko.openstack import topology
from tobiko.shell import ip
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


class NetworkNamespaceFixture(tobiko.SharedFixture):

    def __init__(self,
                 network_namespace: str = None,
                 ssh_client: ssh.SSHClientType = None):
        super().__init__()
        self.network_namespace = network_namespace
        self.ssh_client = ssh_client

    def setup_fixture(self):
        errors = []
        for node in topology.list_openstack_nodes():
            try:
                namespace: str = ip.list_network_namespaces(
                    ignore_errors=True,
                    ssh_client=node.ssh_client).first
            except tobiko.ObjectNotFound:
                LOG.debug(f'No such namespace on host {node.name}',
                          exc_info=1)
            except Exception:
                LOG.debug(f'Error listing namespace on host {node.name}',
                          exc_info=1)
                errors.append(tobiko.exc_info())
            else:
                self.network_namespace = namespace
                self.ssh_client = node.ssh_client
                break
        else:
            if errors:
                raise testtools.MultipleExceptions(*errors)
            else:
                tobiko.skip_test(reason='Network namespace not found')



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.podman import _exception
from tobiko.shell import sh


def discover_podman_socket(ssh_client=None, **execute_params):
    cmd = "systemctl list-sockets | grep podman | awk '{print $1}'"
    result = sh.execute(cmd, stdin=False, stdout=True, stderr=True,
                        expect_exit_status=None, ssh_client=ssh_client,
                        **execute_params)
    if result.exit_status or not result.stdout:
        raise _exception.PodmanSocketNotFoundError(details=result.stderr)
    try:
        socket = result.stdout.splitlines()[0]
    except IndexError as ex:
        podman_error = _exception.PodmanSocketNotFoundError(
            details=result.stderr)
        raise podman_error from ex
    if '0 sockets listed' in socket:
        raise _exception.PodmanSocketNotFoundError(details=socket)
    return socket


def is_podman_running(ssh_client=None, **execute_params):
    try:
        discover_podman_socket(ssh_client=ssh_client, **execute_params)
    except _exception.PodmanSocketNotFoundError:
        return False
    else:
        return True



"""Support files for podman API implementation."""

from __future__ import absolute_import

import collections
import datetime
import functools

from dateutil.parser import parse as dateutil_parse

__all__ = [
    'cached_property',
    'datetime_format',
    'datetime_parse',
    'flatten',
    'fold_keys',
]


def cached_property(fn):
    """Decorate property to cache return value."""
    return property(functools.lru_cache(maxsize=8)(fn))


class ConfigDict(collections.UserDict):
    """Silently ignore None values, only take key once."""

    def __init__(self, **kwargs):
        """Construct dictionary."""
        super().__init__(kwargs)

    def __setitem__(self, key, value):
        """Store unique, not None values."""
        if value is None:
            return

        if super().__contains__(key):
            return

        super().__setitem__(key, value)


class FoldedString(collections.UserString):
    """Foldcase sequences value."""

    def __init__(self, seq):
        super().__init__(seq)
        self.data.casefold()


def fold_keys():  # noqa: D202
    """Fold case of dictionary keys."""

    @functools.wraps(fold_keys)
    def wrapped(mapping):
        """Fold case of dictionary keys."""
        return {k.casefold(): v for (k, v) in mapping.items()}

    return wrapped


def datetime_parse(string):
    """Convert timestamps to datetime.

    tzinfo aware, if provided.
    """
    return dateutil_parse(string.upper(), fuzzy=True)


def datetime_format(dt):
    """Format datetime in consistent style."""
    if isinstance(dt, str):
        return datetime_parse(dt).isoformat()

    if isinstance(dt, datetime.datetime):
        return dt.isoformat()

    raise ValueError('Unable to format {}. Type {} not supported.'.format(
        dt, type(dt)))


def flatten(list_, ltypes=(list, tuple)):
    """Flatten lists of list into a list."""
    ltype = type(list_)
    list_ = list(list_)
    i = 0
    while i < len(list_):
        while isinstance(list_[i], ltypes):
            if not list_[i]:
                list_.pop(i)
                i -= 1
                break
            else:
                list_[i:i + 1] = list_[i]
        i += 1
    return ltype(list_)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import abc
import typing
from abc import ABC

import netaddr
from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import glance
from tobiko.openstack import heat
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack.base import _fixture as base_fixture
from tobiko.openstack.stacks import _hot
from tobiko.openstack.stacks import _neutron
from tobiko.shell import curl
from tobiko.shell import ping
from tobiko.shell import sh
from tobiko.shell import ssh


CONF = config.CONF
LOG = log.getLogger(__name__)


class KeyPairStackFixture(heat.HeatStackFixture):
    template = _hot.heat_template_file('nova/key_pair.yaml')

    public_key: str
    private_key: str

    def setup_fixture(self):
        self.read_keys()
        super(KeyPairStackFixture, self).setup_fixture()

    @property
    def key_file(self) -> str:
        return nova.get_key_file()

    def read_keys(self):
        LOG.info(f'Reading file {self.key_file} for key pairs...')
        with open(self.key_file, 'r') as fd:
            self.private_key = as_str(fd.read())
        with open(self.key_file + '.pub', 'r') as fd:
            self.public_key = as_str(fd.read())


class FlavorStackFixture(heat.HeatStackFixture):
    template = _hot.heat_template_file('nova/flavor.yaml')

    disk: typing.Optional[int] = None
    ephemeral = None
    extra_specs = None
    is_public = None
    name = None
    rxtx_factor = None
    swap: typing.Optional[int] = None
    vcpus = None


@neutron.skip_if_missing_networking_extensions('port-security')
class ServerStackFixture(heat.HeatStackFixture, abc.ABC):

    #: Heat template file
    template = _hot.heat_template_file('nova/server.yaml')

    #: stack with the key pair for the server instance
    key_pair_stack = tobiko.required_fixture(KeyPairStackFixture)

    @property
    def key_file(self) -> str:
        return self.key_pair_stack.key_file

    #: stack with the internal where the server port is created
    network_stack = tobiko.required_fixture(_neutron.NetworkStackFixture)

    #: whenever the server relies only on DHCP for address assignation
    @property
    def need_dhcp(self) -> bool:
        return not self.config_drive

    #: whenever the server will use config-drive to get metadata
    config_drive = False

    image_fixture: tobiko.RequiredFixture[glance.GlanceImageFixture]

    def delete_stack(self, stack_id=None):
        if self._outputs:
            tobiko.cleanup_fixture(self.ssh_client)
        super(ServerStackFixture, self).delete_stack(stack_id=stack_id)

    @property
    def image(self) -> str:
        return self.image_fixture.image_id

    @property
    def username(self) -> str:
        """username used to login to a Nova server instance"""
        return self.image_fixture.username or 'root'

    @property
    def password(self) -> typing.Optional[str]:
        """password used to login to a Nova server instance"""
        return self.image_fixture.password

    @property
    def connection_timeout(self) -> tobiko.Seconds:
        return self.image_fixture.connection_timeout

    @property
    def disabled_algorithms(self) -> typing.Dict[str, typing.Any]:
        return self.image_fixture.disabled_algorithms

    flavor_stack: tobiko.RequiredFixture[FlavorStackFixture]

    @property
    def flavor(self) -> str:
        """Flavor for Nova server instance"""
        return self.flavor_stack.flavor_id

    @property
    def port_security_enabled(self) -> bool:
        """Whenever port security on internal network is enabled"""
        return bool(self.security_groups)

    @property
    def security_groups(self) -> typing.List[str]:
        """Security groups to be associated to network ports"""
        return []

    @property
    def key_name(self) -> str:
        return self.key_pair_stack.key_name

    @property
    def network(self) -> str:
        return self.network_stack.network_id

    #: Floating IP network where the Neutron floating IP are created
    @property
    def floating_network(self) -> str:
        return self.network_stack.floating_network

    @property
    def has_floating_ip(self) -> bool:
        return bool(self.floating_network)

    @property
    def ssh_client_parameters(self) -> typing.Dict[str, typing.Any]:
        return dict(host=self.ip_address,
                    username=self.username,
                    password=self.password,
                    key_filename=self.key_file,
                    connection_timeout=self.connection_timeout,
                    disabled_algorithms=self.disabled_algorithms)

    @property
    def ssh_client(self) -> ssh.SSHClientFixture:
        return ssh.ssh_client(**self.ssh_client_parameters)

    @property
    def peer_ssh_client(self) -> ssh.SSHClientType:
        """Nearest SSH client to an host that can see server fixed IPs ports

        """
        return self.ssh_client

    @property
    def connection(self) -> sh.SSHShellConnection:
        return sh.ssh_shell_connection(self.ssh_client)

    @property
    def ssh_command(self) -> sh.ShellCommand:
        return ssh.ssh_command(host=self.ip_address,
                               username=self.username)

    @property
    def ip_address(self) -> str:
        if self.has_floating_ip:
            return self.floating_ip_address
        else:
            return self.outputs.fixed_ips[0]['ip_address']

    def list_fixed_ips(self, ip_version: typing.Optional[int] = None) -> \
            tobiko.Selection[netaddr.IPAddress]:
        fixed_ips: tobiko.Selection[netaddr.IPAddress] = tobiko.Selection(
            netaddr.IPAddress(fixed_ip['ip_address'])
            for fixed_ip in self.fixed_ips)
        if ip_version is not None:
            fixed_ips.with_attributes(version=ip_version)
        return fixed_ips

    def find_fixed_ip(self,
                      ip_version: int = None,
                      unique=False) -> netaddr.IPAddress:
        fixed_ips = self.list_fixed_ips(ip_version=ip_version)
        if unique:
            return fixed_ips.unique
        else:
            return fixed_ips.first

    @property
    def fixed_ipv4(self):
        return self.find_fixed_ip(ip_version=4)

    @property
    def fixed_ipv6(self):
        return self.find_fixed_ip(ip_version=6)

    @property
    def has_vlan(self) -> bool:
        return False

    #: Schedule on different host that this Nova server instance ID
    different_host = None

    #: Schedule on same host as this Nova server instance ID
    same_host = None

    #: Scheduler group in which this Nova server is attached
    @property
    def scheduler_group(self):
        return None

    @property
    def scheduler_hints(self):
        scheduler_hints = {}
        if self.different_host:
            scheduler_hints.update(different_host=list(self.different_host))
        if self.same_host:
            scheduler_hints.update(same_host=list(self.same_host))
        if self.scheduler_group:
            scheduler_hints.update(group=self.scheduler_group)
        return scheduler_hints

    #: allow to retry creating server in case scheduler hits are not respected
    retry_create = 3
    expected_creted_status = {heat.CREATE_COMPLETE}

    def validate_created_stack(self):
        stack = super(ServerStackFixture, self).validate_created_stack()
        self.validate_scheduler_hints()
        return stack

    @property
    def hypervisor_hostname(self):
        return nova.get_server_hypervisor(server=self.server_details)

    def validate_scheduler_hints(self):
        if self.scheduler_hints:
            hypervisor = self.hypervisor_hostname
            try:
                self.validate_same_host_scheduler_hints(hypervisor=hypervisor)
                self.validate_different_host_scheduler_hints(
                    hypervisor=hypervisor)
            except nova.MigrateServerError as ex:
                raise base_fixture.InvalidFixtureError(
                    name=self.stack_name) from ex

    def validate_same_host_scheduler_hints(self, hypervisor):
        if self.same_host:
            hypervisors = nova.list_servers_hypervisors(self.same_host)
            if hypervisor not in hypervisors:
                self.migrate_server(live=True,
                                    host=hypervisors.unique,
                                    wait_for_guest_os=True)

    def validate_different_host_scheduler_hints(self, hypervisor):
        if self.different_host:
            hypervisors = nova.list_servers_hypervisors(self.different_host)
            if hypervisor in hypervisors:
                self.migrate_server(live=True, wait_for_guest_os=True)

    def wait_for_guest_os_ready(self, timeout=None):
        nova.wait_for_guest_os_ready(server_id=self.server_id,
                                     timeout=timeout)

    def migrate_server(self,
                       live=False,
                       host: str = None,
                       block_migration: bool = None,
                       wait_for_guest_os: bool = False) \
            -> nova.NovaServer:
        server = nova.activate_server(server=self.server_id)
        if wait_for_guest_os:
            self.wait_for_guest_os_ready()
        if live:
            nova.live_migrate_server(server,
                                     host=host,
                                     block_migration=block_migration)
            server = nova.wait_for_server_status(
                server, 'ACTIVE', transient_status=['MIGRATING'])
        else:
            nova.migrate_server(server, host=host)
            server = nova.wait_for_server_status(server, 'VERIFY_RESIZE')
            nova.confirm_resize(server)
            server = nova.wait_for_server_status(
                server, 'ACTIVE', transient_status=['VERIFY_RESIZE'])
        return server

    @property
    def server_details(self) -> nova.NovaServer:
        return nova.get_server(self.server_id)

    @property
    def port_details(self):
        return neutron.get_port(self.port_id)

    def getDetails(self):
        # pylint: disable=W0212
        details = super(ServerStackFixture, self).getDetails()
        stack = self.get_stack()
        if stack:
            details[self.fixture_name + '.stack'] = (
                self.details_content(get_json=lambda: stack._info))
            if stack.stack_status == 'CREATE_COMPLETE':
                details[self.fixture_name + '.server_details'] = (
                    self.details_content(
                        get_json=lambda: self.server_details._info))
                details[self.fixture_name + '.console_output'] = (
                    self.details_content(
                        get_text=lambda: self.console_output))
        return details

    def details_content(self, **kwargs):
        return tobiko.details_content(content_id=self.fixture_name, **kwargs)

    max_console_output_length = 64 * 1024

    @property
    def console_output(self):
        return nova.get_console_output(server=self.server_id,
                                       length=self.max_console_output_length)

    def ensure_server_status(
            self,
            status: str,
            retry_count: int = None,
            retry_timeout: tobiko.Seconds = None,
            retry_interval: tobiko.Seconds = None):
        self.ssh_client.close()
        for attempt in tobiko.retry(count=retry_count,
                                    timeout=retry_timeout,
                                    interval=retry_interval,
                                    default_count=3,
                                    default_timeout=900.,
                                    default_interval=5.):
            tobiko.setup_fixture(self)
            server_id = self.server_id
            try:
                server = nova.ensure_server_status(
                    server=server_id,
                    status=status,
                    timeout=attempt.time_left)
            except nova.WaitForServerStatusError:
                attempt.check_limits()
                LOG.warning(
                    f"Unable to change server '{server_id}' status to "
                    f"'{status}'",
                    exc_info=1)
                tobiko.cleanup_fixture(self)
            else:
                assert server.status == status
                break
        else:
            raise RuntimeError('Broken retry loop')

        return server

    @property
    def nova_required_quota_set(self) -> typing.Dict[str, int]:
        requirements = super().nova_required_quota_set
        requirements['instances'] += 1
        requirements['cores'] += (self.flavor_stack.vcpus or 1)
        return requirements

    @property
    def neutron_required_quota_set(self) -> typing.Dict[str, int]:
        requirements = super().neutron_required_quota_set
        requirements['port'] += 1
        return requirements

    @property
    def is_reachable_timeout(self) -> tobiko.Seconds:
        return None

    def assert_is_reachable(self,
                            ssh_client: ssh.SSHClientType = None,
                            timeout: tobiko.Seconds = None):
        if timeout is None:
            timeout = self.is_reachable_timeout
        ping.assert_reachable_hosts([self.ip_address],
                                    ssh_client=ssh_client,
                                    timeout=timeout)

    def assert_is_unreachable(self,
                              ssh_client: ssh.SSHClientType = None,
                              timeout: tobiko.Seconds = None):
        ping.assert_unreachable_hosts([self.ip_address],
                                      ssh_client=ssh_client,
                                      timeout=timeout)

    @property
    def floating_ip_details(self):
        return neutron.get_floating_ip(self.floating_ip_id)

    user_data = None


class CloudInitServerStackFixture(ServerStackFixture, ABC):

    #: SWAP file name
    swap_filename: str = '/swap.img'
    #: SWAP file size in bytes
    swap_size: typing.Optional[int] = None
    #: nax SWAP file size in bytes
    swap_maxsize: typing.Optional[int] = None

    @tobiko.interworker_synched('cloudinit_server_setup_fixture')
    def setup_fixture(self):
        super(CloudInitServerStackFixture, self).setup_fixture()

        if config.get_bool_env('TOBIKO_PREVENT_CREATE'):
            LOG.debug("skip wait_for_guest_os_ready during check-resources "
                      "steps because the console output may be empty in some "
                      "cases, such as hypervisor reboot")
        else:
            self.wait_for_guest_os_ready()

        if self.has_floating_ip:
            self.assert_is_reachable()
            self.wait_for_cloud_init_done()

    def wait_for_guest_os_ready(self, timeout=900.):
        super(CloudInitServerStackFixture, self).wait_for_guest_os_ready(
            timeout=timeout)

    @property
    def is_reachable_timeout(self) -> tobiko.Seconds:
        # I expect cloud-init based servers to be slow to boot
        return CONF.tobiko.nova.cloudinit_is_reachable_timeout

    @property
    def user_data(self):
        return nova.user_data(self.cloud_config)

    @property
    def cloud_config(self):
        cloud_config = nova.cloud_config()
        # default is to not create any swap files,
        # because 'swap_file_max_size' is set to None
        if self.swap_maxsize is not None:
            cloud_config = nova.cloud_config(
                cloud_config,
                swap={'filename': self.swap_filename,
                      'size': self.swap_size or 'auto',
                      'maxsize': self.swap_maxsize})
        return cloud_config

    def wait_for_cloud_init_done(self, **params):
        nova.wait_for_cloud_init_done(ssh_client=self.ssh_client,
                                      timeout=self.is_reachable_timeout,
                                      **params)


class ExternalServerStackFixture(ServerStackFixture, abc.ABC):
    # pylint: disable=abstract-method

    #: stack with the network where the server port is created
    network_stack = tobiko.required_fixture(
        _neutron.ExternalNetworkStackFixture)

    # external servers doesn't need floating IPs
    has_floating_ip = False

    # We must rely on ways of configuring IPs without relying on DHCP
    config_drive = True

    # external network servers are visible from test host
    peer_ssh_client = None

    # external network DHCP could conflict with Neutron one
    need_dhcp = False

    @property
    def floating_network(self):
        return self.network_stack.network_id


class PeerServerStackFixture(ServerStackFixture, abc.ABC):
    """Server witch networking access requires passing by another Nova server
    """

    has_floating_ip = False

    peer_stack: tobiko.RequiredFixture[ServerStackFixture]

    @property
    def ssh_client_parameters(self) -> typing.Dict[str, typing.Any]:
        parameters = super().ssh_client_parameters
        parameters.update(proxy_jump=self.peer_ssh_client)
        return parameters

    @property
    def peer_ssh_client(self) -> ssh.SSHClientFixture:
        return self.peer_stack.ssh_client

    @property
    def ssh_command(self) -> sh.ShellCommand:
        proxy_command = self.peer_stack.ssh_command + [
            'nc', self.ip_address, '22']
        return ssh.ssh_command(host=self.ip_address,
                               username=self.username,
                               proxy_command=proxy_command)

    @property
    def network(self) -> str:
        return self.peer_stack.network


@nova.skip_if_missing_hypervisors(count=2, state='up', status='enabled')
class DifferentHostServerStackFixture(PeerServerStackFixture, abc.ABC):
    # pylint: disable=abstract-method

    @property
    def different_host(self):
        return [self.peer_stack.server_id]


class SameHostServerStackFixture(PeerServerStackFixture, abc.ABC):

    @property
    def same_host(self):
        return [self.peer_stack.server_id]


def as_str(text):
    if isinstance(text, str):
        return text
    else:
        return text.decode()


class HttpServerStackFixture(PeerServerStackFixture, abc.ABC):

    http_server_port = 80

    http_request_scheme = 'http'
    http_request_path = ''

    def send_http_request(
            self,
            hostname: typing.Union[str, netaddr.IPAddress, None] = None,
            ip_version: typing.Optional[int] = None,
            port: typing.Optional[int] = None,
            path: typing.Optional[str] = None,
            retry_count: typing.Optional[int] = None,
            retry_timeout: tobiko.Seconds = None,
            retry_interval: tobiko.Seconds = None,
            ssh_client: typing.Optional[ssh.SSHClientFixture] = None,
            **curl_parameters) -> str:
        if hostname is None:
            hostname = self.find_fixed_ip(ip_version=ip_version)
        if port is None:
            port = self.http_server_port
        if path is None:
            path = self.http_request_path
        if ssh_client is None:
            ssh_client = self.peer_stack.ssh_client
        return curl.execute_curl(scheme='http',
                                 hostname=hostname,
                                 port=port,
                                 path=path,
                                 retry_count=retry_count,
                                 retry_timeout=retry_timeout,
                                 retry_interval=retry_interval,
                                 ssh_client=ssh_client,
                                 **curl_parameters)


class ServerGroupStackFixture(heat.HeatStackFixture):
    template = _hot.heat_template_file('nova/server_group.yaml')


class AffinityServerGroupStackFixture(tobiko.SharedFixture):
    server_group_stack = tobiko.required_fixture(
        ServerGroupStackFixture)

    @property
    def scheduler_group(self):
        return self.server_group_stack.affinity_server_group_id


class AntiAffinityServerGroupStackFixture(tobiko.SharedFixture):
    server_group_stack = tobiko.required_fixture(
        ServerGroupStackFixture)

    @property
    def scheduler_group(self):
        return self.server_group_stack.anti_affinity_server_group_id



# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import stacks


@keystone.skip_unless_has_keystone_credentials()
class RouterTest(testtools.TestCase):
    """Tests network creation"""

    #: Stack of resources with a network with a gateway router
    stack = tobiko.required_fixture(stacks.NetworkStackFixture)

    def test_get_router_with_id(self):
        if not self.stack.has_gateway:
            tobiko.skip_test(f"Stack {self.stack.stack_name} has no gateway "
                             "router")
        router = neutron.get_router(self.stack.gateway_id)
        self.assertEqual(self.stack.gateway_id, router['id'])

    def test_get_router_with_details(self):
        if not self.stack.has_gateway:
            tobiko.skip_test(f"Stack {self.stack.stack_name} has no gateway "
                             "router")
        router = neutron.get_router(self.stack.gateway_details)
        self.assertEqual(self.stack.gateway_id, router['id'])

    def test_create_router(self,
                           network: neutron.NetworkIdType = None) \
            -> neutron.PortType:
        router = neutron.create_router(name=self.id(), network=network)
        self.assertIsInstance(router['id'], str)
        self.assertNotEqual('', router['id'])
        self.assertEqual(self.id(), router['name'])
        self.assertEqual(router['id'], neutron.get_router(router)['id'])
        if network is not None:
            self.assertEqual(neutron.get_network_id(network),
                             router['external_gateway_info']['network_id'])
        return router

    def test_create_router_with_network_id(self):
        network_id = self.stack.floating_network
        self.test_create_router(network=network_id)

    def test_create_router_with_network_details(self):
        network = {'id': self.stack.floating_network}
        self.test_create_router(network=network)

    def test_delete_router_with_id(self):
        router = self.test_create_router()
        neutron.delete_router(router['id'])
        self.assertRaises(neutron.NoSuchRouter, neutron.get_router,
                          router['id'])

    def test_delete_router_with_details(self):
        router = self.test_create_router()
        neutron.delete_router(router)
        self.assertRaises(neutron.NoSuchRouter, neutron.get_router,
                          router['id'])

    def test_delete_router_with_invalid_id(self):
        self.assertRaises(neutron.NoSuchRouter,
                          neutron.delete_router, '<invalid-id>')

    def test_add_router_interface_with_port(self):
        port = neutron.create_port(name=self.id())
        router = self.test_create_router()
        interface = neutron.add_router_interface(router=router, port=port)
        self.assertEqual(port['network_id'], interface['network_id'])
        return interface

    def test_remove_router_interface_with_port(self):
        interface = self.test_add_router_interface_with_port()
        neutron.remove_router_interface(router=interface['id'],
                                        port=interface['port_id'])

    def test_add_router_interface_with_subnet(self):
        network = neutron.create_network(name=self.id())
        subnet_pool_range = "172.168.111.0/24"
        subnet_pool_default_prefixlen = 26
        subnet_pool = neutron.create_subnet_pool(
            name=self.id(),
            prefixes=[subnet_pool_range],
            default_prefixlen=subnet_pool_default_prefixlen)
        subnet = neutron.create_subnet(name=self.id(),
                                       network=network,
                                       subnetpool_id=subnet_pool['id'],
                                       ip_version=4)
        router = self.test_create_router()
        interface = neutron.add_router_interface(router=router, subnet=subnet)
        self.assertEqual(subnet['id'], interface['subnet_id'])
        return interface

    def test_remove_router_interface_with_subnet(self):
        interface = self.test_add_router_interface_with_subnet()
        neutron.remove_router_interface(router=interface['id'],
                                        subnet=interface['subnet_id'])



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import
from __future__ import division

import dpkt
from oslo_log import log

import tobiko


LOG = log.getLogger(__name__)


def assert_pcap_content(pcap: dpkt.pcap.Reader, expect_empty: bool):
    actual_empty = True
    for _ in pcap:
        actual_empty = False
        break
    testcase = tobiko.get_test_case()
    LOG.debug(f'Is the obtained pcap file empty? {actual_empty}')
    testcase.assertEqual(expect_empty, actual_empty)


def assert_pcap_is_empty(pcap: dpkt.pcap.Reader):
    LOG.debug('This test expects an empty pcap capture')
    assert_pcap_content(pcap, True)


def assert_pcap_is_not_empty(pcap: dpkt.pcap.Reader):
    LOG.debug('This test expects a non-empty pcap capture')
    assert_pcap_content(pcap, False)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import json
import typing

import netaddr
from neutronclient.common import exceptions as nc_exceptions
from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import heat
from tobiko.openstack import neutron
from tobiko.openstack.base import _fixture as base_fixture
from tobiko.openstack.stacks import _hot
from tobiko.shell import ip
from tobiko.shell import sh
from tobiko.shell import ssh


CONF = config.CONF
LOG = log.getLogger(__name__)


class ExternalNetworkStackFixture(heat.HeatStackFixture):

    template = _hot.heat_template_file('neutron/external_network.yaml')

    @tobiko.interworker_synched('create_external_network_stack')
    def setup_stack(self):
        super().setup_stack()

    @property
    def external_name(self) -> typing.Optional[str]:
        return tobiko.tobiko_config().neutron.external_network

    _external_network: typing.Optional[neutron.NetworkType] = None

    @property
    def external_network(self) -> typing.Optional[neutron.NetworkType]:
        external_network = self._external_network
        if external_network is None:
            for network in list_external_networks(name=self.external_name):
                if not network['subnets']:
                    LOG.debug(f"Network '{network['id']}' has no subnet")
                    continue

                network_dump = json.dumps(network, indent=4, sort_keys=True)
                LOG.debug(f"Found external network for {self.fixture_name}:\n"
                          f"{network_dump}")
                self._external_network = external_network = network
                break
            else:
                LOG.warning("No external network found for "
                            f"'{self.fixture_name}':\n"
                            f" - name or ID: {self.external_name}\n")
        return external_network

    @property
    def external_id(self) -> typing.Optional[str]:
        network = self.external_network
        if network is None:
            return None
        else:
            return network['id']

    @property
    def has_external_id(self):
        return bool(self.external_network)

    @property
    def network_details(self) -> neutron.NetworkType:
        return neutron.get_network(self.network_id)

    @property
    def has_network(self) -> bool:
        return bool(self.network_id)

    has_gateway = False

    @property
    def create_router(self) -> bool:
        return False

    @property
    def has_l3_ha(self):
        """Whenever can obtain gateway router HA value"""
        return neutron.has_networking_extensions('l3-ha')

    @property
    def has_dvr(self):
        """Whenever to require a distributed router"""
        return neutron.has_networking_extensions('dvr')

    ha = False
    distributed: typing.Optional[bool] = None

    @property
    def router_value_specs(self) -> typing.Dict[str, typing.Any]:
        value_specs: typing.Dict[str, typing.Any] = {}
        if self.has_l3_ha and not neutron.has_ovn():
            value_specs.update(ha=bool(self.ha))
        if self.distributed is not None and self.has_dvr:
            value_specs.update(distributed=bool(self.distributed))
        return value_specs

    @property
    def has_router(self) -> bool:
        return bool(self.router_id)

    @property
    def router_details(self) -> neutron.RouterType:
        return neutron.get_router(self.router_id)


class RouterStackFixture(ExternalNetworkStackFixture):

    def __init__(self,
                 neutron_client: neutron.NeutronClientType = None):
        super(RouterStackFixture, self).__init__()
        self._neutron_client = neutron_client

    distributed: typing.Optional[bool] = None

    @property
    def create_router(self) -> bool:
        return self.has_external_id

    @property
    def neutron_required_quota_set(self) -> typing.Dict[str, int]:
        requirements = super().neutron_required_quota_set
        if self.create_router:
            requirements['router'] += 1
        return requirements

    def ensure_router_interface(self,
                                subnet: neutron.SubnetIdType = None,
                                network: neutron.NetworkIdType = None):
        ensure_router_interface(
            network=network,
            subnet=subnet,
            router=self.router_id,
            client=self.neutron_client,
            create_router_interface_func=self.create_router_interface)

    @property
    def neutron_client(self) -> neutron.NeutronClientType:
        if self._neutron_client is None:
            self._neutron_client = neutron.neutron_client()
        return self._neutron_client

    @staticmethod
    def create_router_interface(
            router: neutron.RouterIdType = None,
            subnet: neutron.SubnetIdType = None,
            network: neutron.NetworkIdType = None,
            client: neutron.NeutronClientType = None,
            add_cleanup=False) -> neutron.PortType:
        return create_router_interface(router=router,
                                       subnet=subnet,
                                       network=network,
                                       client=client,
                                       add_cleanup=add_cleanup)


def create_router_interface(
        router: neutron.RouterIdType = None,
        subnet: neutron.SubnetIdType = None,
        network: neutron.NetworkIdType = None,
        client: neutron.NeutronClientType = None,
        add_cleanup=False) \
        -> neutron.PortType:
    stack = RouterInterfaceStackFixture(router=router,
                                        subnet=subnet,
                                        network=network,
                                        neutron_client=client)
    if add_cleanup:
        tobiko.use_fixture(stack)
    else:
        tobiko.setup_fixture(stack)
    return stack.port_details


def ensure_router_interface(
        router: neutron.RouterIdType = None,
        subnet: neutron.SubnetIdType = None,
        network: neutron.NetworkIdType = None,
        client: neutron.NeutronClientType = None,
        add_cleanup=False,
        create_router_interface_func=create_router_interface) \
        -> neutron.PortType:
    client = neutron.neutron_client(client)
    if router is None:
        router = get_router_id()
    if subnet is None and network is None:
        raise ValueError('Must specify a network or a subnet')
    try:
        port = neutron.find_port(network=network,
                                 device=router,
                                 subnet=subnet,
                                 client=client)
    except tobiko.ObjectNotFound:
        pass
    else:
        port_dump = json.dumps(port, indent=4, sort_keys=True)
        LOG.debug(f'Router interface already exist:\n{port_dump}')
        return port

    if subnet is None:
        assert network is not None
        LOG.info("Add router interface to network "
                 f"{neutron.get_network_id(network)}")
        for _subnet in neutron.list_subnets(network=network,
                                            client=client):
            neutron.ensure_subnet_gateway(subnet=_subnet,
                                          client=client)
    else:
        subnet = neutron.ensure_subnet_gateway(subnet=subnet,
                                               client=client)
        gateway_ip = subnet['gateway_ip']
        try:
            port = neutron.find_port(fixed_ips=[f'ip_address={gateway_ip}'],
                                     client=client)
        except tobiko.ObjectNotFound:
            LOG.info("Add router interface to subnet "
                     f"{neutron.get_subnet_id(subnet)}")
        else:
            # it must force router to bind new network port because subnet
            # gateway IP address is already being used
            port_dump = json.dumps(port, indent=4, sort_keys=True)
            LOG.debug(f'Port with gateway IP already exists:\n{port_dump}')
            if network is None:
                network = subnet['network_id']
            subnet = None
            LOG.info("Add router interface to network "
                     f"{neutron.get_network_id(network)}")
    return create_router_interface_func(router=router,
                                        subnet=subnet,
                                        network=network,
                                        client=client,
                                        add_cleanup=add_cleanup)


class RouterNoSnatStackFixture(RouterStackFixture):
    """Extra Router configured with SNAT disabled"""
    def setup_fixture(self):
        super().setup_fixture()
        egi = neutron.get_router(self.router_id)['external_gateway_info']
        egi['enable_snat'] = False
        neutron.update_router(self.router_id, external_gateway_info=egi)


@neutron.skip_if_missing_networking_extensions('subnet_allocation')
class SubnetPoolFixture(base_fixture.ResourceFixture):
    """Neutron Subnet Pool Fixture.

    A subnet pool is a dependency of network fixtures with either IPv4 or
    IPv6 subnets (or both). The CIDRs for those subnets are obtained from this
    resource.
    NOTE: this fixture does not represent a heat stack, but it is under the
    stacks module until a decision is taken on where this kind of fixtures are
    located
    """

    _resource: typing.Optional[neutron.SubnetPoolType] = None
    prefixes: list = [CONF.tobiko.neutron.ipv4_cidr]
    default_prefixlen: int = CONF.tobiko.neutron.ipv4_prefixlen
    _not_found_exception_tuple: typing.Type[
        neutron.NoSuchSubnetPool] = (neutron.NoSuchSubnetPool)

    def __init__(self, prefixes=None, default_prefixlen=None):
        super().__init__()
        self.prefixes = prefixes or self.prefixes
        self.default_prefixlen = default_prefixlen or self.default_prefixlen

    @property
    def subnet_pool_id(self):
        return self.resource_id

    @property
    def subnet_pool(self):
        return self.resource

    @tobiko.interworker_synched('create_subnet_pool')
    def try_create_resource(self):
        super().try_create_resource()

    def resource_create(self):
        return neutron.create_subnet_pool(
            name=self.name, prefixes=self.prefixes,
            default_prefixlen=self.default_prefixlen, add_cleanup=False)

    def resource_delete(self):
        neutron.delete_subnet_pool(self.subnet_pool_id)

    def resource_find(self):
        return neutron.find_subnet_pool(name=self.name)


class SubnetPoolIPv6Fixture(SubnetPoolFixture):
    prefixes: list = [CONF.tobiko.neutron.ipv6_cidr]
    default_prefixlen: int = CONF.tobiko.neutron.ipv6_prefixlen


@neutron.skip_if_missing_networking_extensions('port-security')
class NetworkBaseStackFixture(heat.HeatStackFixture):
    """Heat stack for creating internal network with a router to external"""
    subnet_pools_ipv4_stack: typing.Optional[tobiko.RequiredFixture] = (
        tobiko.required_fixture(SubnetPoolFixture)
        if bool(CONF.tobiko.neutron.ipv4_cidr)
        else None)
    subnet_pools_ipv6_stack: typing.Optional[tobiko.RequiredFixture] = (
        tobiko.required_fixture(SubnetPoolIPv6Fixture)
        if bool(CONF.tobiko.neutron.ipv6_cidr)
        else None)

    #: Heat template file
    template = _hot.heat_template_file('neutron/network.yaml')

    #: Enable port security by default for new network ports
    port_security_enabled = True

    @property
    def has_ipv4(self):
        """Whenever to setup IPv4 subnet"""
        return bool(CONF.tobiko.neutron.ipv4_cidr)

    @property
    def has_ipv6(self):
        """Whenever to setup IPv6 subnet"""
        return bool(CONF.tobiko.neutron.ipv6_cidr)

    @property
    def subnet_pool_ipv4_id(self):
        return self.subnet_pools_ipv4_stack.subnet_pool_id

    @property
    def subnet_pool_ipv6_id(self):
        return self.subnet_pools_ipv6_stack.subnet_pool_id

    @property
    def network_value_specs(self):
        """Extra network creation parameters"""
        return {}

    gateway_stack = tobiko.required_fixture(RouterStackFixture)

    @property
    def ha(self) -> bool:
        return self.gateway_stack.ha

    @property
    def floating_network(self):
        """Network ID where the Neutron floating IPs are created"""
        return self.gateway_stack.network_id

    @property
    def gateway_network(self):
        """Network ID where gateway routes packages to"""
        return self.floating_network

    @property
    def gateway(self) -> str:
        return self.gateway_stack.router_id

    @property
    def has_gateway(self):
        """Whenever to setup gateway router"""
        return bool(self.gateway_network)

    @property
    def has_net_mtu(self):
        """Whenever can obtain network MTU value"""
        return neutron.has_networking_extensions('net-mtu')

    @property
    def has_l3_ha(self):
        """Whenever can obtain gateway router HA value"""
        return neutron.has_networking_extensions('l3-ha')

    @property
    def network_details(self):
        return neutron.get_network(self.network_id)

    @property
    def ipv4_subnet_details(self):
        return neutron.get_subnet(self.ipv4_subnet_id)

    @property
    def ipv4_subnet_cidr(self):
        return netaddr.IPNetwork(self.ipv4_subnet_details['cidr'])

    @property
    def ipv4_subnet_gateway_ip(self):
        return netaddr.IPAddress(self.ipv4_subnet_details['gateway_ip'])

    @property
    def ipv4_dns_nameservers(self):
        nameservers = CONF.tobiko.neutron.ipv4_dns_nameservers
        if nameservers is None:
            nameservers = default_nameservers(ip_version=4)
        return ','.join(str(nameserver) for nameserver in nameservers)

    @property
    def ipv6_subnet_details(self):
        return neutron.get_subnet(self.ipv6_subnet_id)

    @property
    def ipv6_subnet_cidr(self):
        return netaddr.IPNetwork(self.ipv6_subnet_details['cidr'])

    @property
    def ipv6_subnet_gateway_ip(self):
        return netaddr.IPAddress(self.ipv6_subnet_details['gateway_ip'])

    @property
    def ipv6_dns_nameservers(self):
        nameservers = CONF.tobiko.neutron.ipv6_dns_nameservers
        if nameservers is None:
            nameservers = default_nameservers(ip_version=6)
        return ','.join(str(nameserver) for nameserver in nameservers)

    @property
    def gateway_details(self):
        return neutron.get_router(self.gateway_id)

    @property
    def external_gateway_ips(self):
        fixed_ips = self.gateway_details['external_gateway_info'][
            'external_fixed_ips']
        return tobiko.select(netaddr.IPAddress(fixed_ip['ip_address'])
                             for fixed_ip in fixed_ips)

    @property
    def ipv4_gateway_ports(self):
        return neutron.list_ports(fixed_ips='subnet_id=' + self.ipv4_subnet_id,
                                  device_id=self.gateway_id,
                                  network_id=self.network_id)

    @property
    def ipv6_gateway_ports(self):
        return neutron.list_ports(fixed_ips='subnet_id=' + self.ipv6_subnet_id,
                                  device_id=self.gateway_id,
                                  network_id=self.network_id)

    @property
    def external_geteway_ports(self):
        return neutron.list_ports(device_id=self.gateway_id,
                                  network_id=self.gateway_network_id)

    @property
    def ipv4_gateway_addresses(self):
        ips = tobiko.Selection()
        for port in self.ipv4_gateway_ports:
            ips.extend(neutron.list_port_ip_addresses(port))
        return ips

    @property
    def ipv6_gateway_addresses(self):
        ips = tobiko.Selection()
        for port in self.ipv6_gateway_ports:
            ips.extend(neutron.list_port_ip_addresses(port))
        return ips

    @property
    def external_gateway_addresses(self):
        ips = tobiko.Selection()
        for port in self.external_geteway_ports:
            ips.extend(neutron.list_port_ip_addresses(port))
        return ips

    @property
    def gateway_network_details(self):
        return neutron.get_network(self.gateway_network_id)

    @property
    def neutron_required_quota_set(self) -> typing.Dict[str, int]:
        requirements = super().neutron_required_quota_set
        requirements['network'] += 1
        if self.has_ipv4:
            requirements['subnet'] += 1
        if self.has_ipv6:
            requirements['subnet'] += 1
        return requirements

    def is_router_distributed(self) -> bool:
        if self.has_gateway:
            tobiko.setup_fixture(self)
            return bool(self.gateway_details.get('distributed'))
        else:
            return False

    @classmethod
    def skip_if_router_is_distributed(cls, reason: str = None):
        fixture = tobiko.get_fixture(cls)
        if reason is None:
            reason = "Distributed router is not supported"
        return tobiko.skip_if(reason=reason,
                              predicate=fixture.is_router_distributed)


class NetworkStackFixture(NetworkBaseStackFixture):
    @tobiko.interworker_synched('create_network_stack')
    def setup_stack(self):
        super().setup_stack()


class NetworkNoFipStackFixture(NetworkBaseStackFixture):
    """Extra Network Stack where VMs will not have FIPs"""
    gateway_stack = tobiko.required_fixture(RouterNoSnatStackFixture)

    @tobiko.interworker_synched('create_network_nofip_stack')
    def setup_stack(self):
        super().setup_stack()


@neutron.skip_if_missing_networking_extensions('net-mtu-writable')
class NetworkWithNetMtuWriteStackFixture(NetworkBaseStackFixture):

    @tobiko.interworker_synched('create_network_withnetmtuwrite_stack')
    def setup_stack(self):
        super().setup_stack()

    @property
    def custom_mtu_size(self):
        return CONF.tobiko.neutron.custom_mtu_size

    @property
    def network_value_specs(self):
        value_specs = super(NetworkWithNetMtuWriteStackFixture,
                            self).network_value_specs
        return dict(value_specs, mtu=self.custom_mtu_size)


@neutron.skip_if_missing_networking_extensions('security-group')
class SecurityGroupsFixture(heat.HeatStackFixture):
    """Heat stack with some security groups

    """
    #: Heat template file
    template = _hot.heat_template_file('neutron/security_groups.yaml')

    def __init__(self, stateful: bool = True):
        self._stateful = stateful
        super(SecurityGroupsFixture, self).__init__()

    @property
    def stateful(self):
        return self._stateful


@neutron.skip_if_missing_networking_extensions('stateful-security-group')
class StatelessSecurityGroupFixture(base_fixture.ResourceFixture):
    """Neutron Stateless Security Group Fixture.

    This SG will by default allow SSH and ICMP to the instance and also
    ingress traffic from the metadata service as it can't rely on conntrack.
    """

    _resource: typing.Optional[neutron.SecurityGroupType] = None
    description: typing.Optional[str] = ""
    rules = [
        {
            'protocol': 'tcp',
            'port_range_min': 22,
            'port_range_max': 22,
            'direction': 'ingress'
        }, {
            'protocol': 'icmp',
            'direction': 'ingress'
        }, {
            'protocol': 'tcp',
            'remote_ip_prefix': '%s/32' % neutron.METADATA_IPv4,
            'direction': 'ingress'
        }
    ]
    _not_found_exception_tuple: typing.Type[nc_exceptions.NotFound] = (
        neutron.NotFound)

    def __init__(self, description=None, rules=None):
        super().__init__()
        self.description = description or self.description
        self.rules = rules or self.rules

    @property
    def neutron_required_quota_set(self) -> typing.Dict[str, int]:
        requirements = super().neutron_required_quota_set
        requirements['security_group'] += 1
        return requirements

    @property
    def security_group_id(self):
        return self.resource_id

    @property
    def security_group(self):
        return self.resource

    @tobiko.interworker_synched('create_security_group')
    def try_create_resource(self):
        super().try_create_resource()

    def resource_create(self):
        sg = neutron.create_security_group(
            name=self.name, description=self.description,
            add_cleanup=False, stateful=False)

        # add rules once the SG was created
        for rule in self.rules:
            neutron.create_security_group_rule(
                sg['id'], add_cleanup=False, **rule)

        # return the updated SG, including the rules just added
        return self.resource_find()

    def resource_delete(self):
        neutron.delete_security_group(self.security_group_id)

    def resource_find(self):
        sgs = neutron.list_security_groups(name=self.name)
        if len(sgs) == 0:
            LOG.debug("Security group %r not found.", self.name)
            return None
        else:
            return sgs.unique


def list_external_networks(name: str = None) -> \
        tobiko.Selection[neutron.NetworkType]:
    networks = tobiko.Selection[neutron.NetworkType]()
    if name is not None:
        try:
            network = neutron.get_network(name)
        except neutron.NoSuchNetwork:
            LOG.error(f"No such network with ID '{name}'")
        else:
            networks.append(network)
    if not networks:
        network_params = {'router:external': True, "status": "ACTIVE"}
        if name is not None:
            network_params['name'] = name
        networks += neutron.list_networks(**network_params)
    if not networks and name:
        raise ValueError("No such external network with name or ID "
                         f"'{name}'")
    return networks


def get_external_network_id():
    return tobiko.setup_fixture(ExternalNetworkStackFixture).network_id


def get_external_network():
    return tobiko.setup_fixture(ExternalNetworkStackFixture).network_details


def has_external_network():
    return tobiko.setup_fixture(ExternalNetworkStackFixture).has_network


skip_unless_has_external_network = tobiko.skip_unless(
    'External network not found', has_external_network)


def get_floating_network_id() -> str:
    return tobiko.setup_fixture(RouterStackFixture).network_id


def get_floating_network() -> neutron.NetworkType:
    return tobiko.setup_fixture(RouterStackFixture).network_details


def has_floating_network() -> bool:
    return tobiko.setup_fixture(RouterStackFixture).has_network


skip_unless_has_floating_network = tobiko.skip_unless(
    'Floating network not found', has_floating_network)


def get_router_id() -> str:
    return tobiko.setup_fixture(RouterStackFixture).router_id


def get_router() -> neutron.RouterType:
    return tobiko.setup_fixture(RouterStackFixture).router_details


def has_router() -> bool:
    return tobiko.setup_fixture(RouterStackFixture).has_router


skip_unless_has_router = tobiko.skip_unless(
    'External router not created', has_router)


class DefaultNameserversFixture(tobiko.SharedFixture):

    remove_local_ips = True
    max_count = 3
    ip_version = None

    nameservers: typing.Optional[tobiko.Selection[netaddr.IPAddress]] = None

    @property
    def ssh_client(self):
        host = tobiko.tobiko_config().neutron.nameservers_host
        if host is None:
            return ssh.ssh_proxy_client()
        else:
            return ssh.ssh_client(host)

    @property
    def filenames(self):
        return tuple(tobiko.tobiko_config().neutron.nameservers_filenames)

    def setup_fixture(self):
        self.nameservers = self.list_nameservers()

    def list_nameservers(self) -> tobiko.Selection[netaddr.IPAddress]:
        nameservers: tobiko.Selection[netaddr.IPAddress]
        if has_external_network():
            network_id = get_external_network_id()
            nameservers = neutron.list_network_nameservers(
                network_id=network_id)
            LOG.debug("Nameservers copied from external network: "
                      f"{nameservers}")
        else:
            # Copy nameservers from target host
            nameservers = sh.list_nameservers(ssh_client=self.ssh_client,
                                              ip_version=self.ip_version,
                                              filenames=self.filenames)
            if self.remove_local_ips:
                local_ips = ip.list_ip_addresses(scope='host')
                if local_ips:
                    # Filter out all local IPs
                    nameservers = tobiko.Selection[netaddr.IPAddress](
                        nameserver for nameserver in nameservers
                        if nameserver not in local_ips)
            LOG.debug(f"Nameservers copied from host: {nameservers}")
        if self.max_count:
            # Keep only up to max_count nameservers
            actual_count = len(nameservers)
            if actual_count > self.max_count:
                LOG.waring("Limit the number of nameservers from "
                           f"{actual_count} to {self.max_count}: "
                           f"{nameservers}")
                nameservers = tobiko.Selection[netaddr.IPAddress](
                    nameservers[:self.max_count])
        return nameservers


DEFAULT_NAMESERVERS_FIXTURE = DefaultNameserversFixture


def default_nameservers(
        ip_version: typing.Optional[int] = None) -> \
        tobiko.Selection[netaddr.IPAddress]:
    nameservers = tobiko.setup_fixture(
        DEFAULT_NAMESERVERS_FIXTURE).nameservers
    if ip_version is not None:
        nameservers = nameservers.with_attributes(version=ip_version)
    return nameservers


class FloatingIpStackFixture(heat.HeatStackFixture):

    #: Heat template file
    template = _hot.heat_template_file('neutron/floating_ip.yaml')

    router_stack = tobiko.required_fixture(RouterStackFixture)

    def __init__(self,
                 stack_name: str = None,
                 network: neutron.NetworkIdType = None,
                 port: neutron.PortIdType = None,
                 device_id: str = None,
                 fixed_ip_address: str = None,
                 heat_client: heat.HeatClientType = None,
                 neutron_client: neutron.NeutronClientType = None):
        self._network = network
        self._port = port
        self._neutron_client = neutron_client
        self._device_id = device_id
        self._fixed_ip_address = fixed_ip_address
        super(FloatingIpStackFixture, self).__init__(stack_name=stack_name,
                                                     client=heat_client)

    @property
    def network(self) -> str:
        network = self._network
        if network is None:
            return self.router_stack.network_id
        else:
            return neutron.get_network_id(network)

    @property
    def port(self) -> str:
        if isinstance(self._port, str):
            return self._port
        else:
            return self.port_details['id']

    @property
    def port_details(self) -> neutron.PortType:
        if self._port is None:
            params: typing.Dict[str, typing.Any] = {}
            device_id = self.device_id
            if device_id is not None:
                params['device_id'] = device_id
            self._port = neutron.find_port(
                client=self.neutron_client,
                fixed_ips=[f'ip_address={self.fixed_ip_address}'],
                **params)
        elif isinstance(self._port, str):
            self._port = neutron.get_port(self.port,
                                          client=self.neutron_client)
        assert isinstance(self._port, dict)
        return self._port

    @property
    def fixed_ip_address(self) -> str:
        if self._fixed_ip_address is None:
            raise ValueError(
                'Must specify at least a port or a fixed IP address')
        return self._fixed_ip_address

    @property
    def device_id(self) -> typing.Optional[str]:
        return self._device_id

    def setup_stack_name(self) -> str:
        stack_name = self.stack_name
        if stack_name is None:
            self.stack_name = stack_name = f"{self.fixture_name}-{self.port}"
        return stack_name

    def prepare_external_resources(self):
        super().prepare_external_resources()
        for fixed_ip in self.port_details['fixed_ips']:
            self.router_stack.ensure_router_interface(
                subnet=fixed_ip['subnet_id'])

    @property
    def network_details(self) -> neutron.NetworkType:
        return neutron.get_network(self.network_id)

    @property
    def router_details(self) -> neutron.RouterType:
        return neutron.get_router(self.router_id)

    @property
    def floating_ip_details(self) -> neutron.FloatingIpType:
        return neutron.get_floating_ip(self.floating_ip_id)

    @property
    def neutron_client(self) -> neutron.NeutronClientType:
        if self._neutron_client is None:
            self._neutron_client = self.router_stack.neutron_client
        return self._neutron_client


class RouterInterfaceStackFixture(heat.HeatStackFixture):

    #: Heat template file
    template = _hot.heat_template_file('neutron/router_interface.yaml')

    def __init__(self,
                 stack_name: str = None,
                 router: neutron.RouterIdType = None,
                 network: neutron.NetworkIdType = None,
                 subnet: neutron.SubnetIdType = None,
                 neutron_client: neutron.NeutronClientType = None):
        self._router = router
        self._network = network
        self._subnet = subnet
        self._neutron_client = neutron_client
        super().__init__(stack_name=stack_name)

    @property
    def router(self) -> str:
        if self._router is None:
            self._router = get_router_id()
        return neutron.get_router_id(self._router)

    @property
    def router_details(self) -> neutron.RouterType:
        if self._router is None or isinstance(self._router, str):
            self._router = neutron.get_router(self.router,
                                              client=self.neutron_client)
        return self._router

    @property
    def network(self) -> str:
        if self._network is None:
            subnet = self.subnet_details
            if subnet is None:
                raise ValueError('Must specify at least network or subnet')
            self._network = subnet['network_id']
        return neutron.get_network_id(self._network)

    @property
    def network_details(self) -> neutron.NetworkType:
        if self._network is None or isinstance(self._network, str):
            self._network = neutron.get_network(self.network,
                                                client=self.neutron_client)
        return self._network

    @property
    def subnet(self) -> typing.Optional[str]:
        if self._subnet is None:
            return None
        else:
            return neutron.get_subnet_id(self._subnet)

    @property
    def subnet_details(self) -> typing.Optional[neutron.SubnetType]:
        if self._subnet is None:
            return None
        if isinstance(self._subnet, str):
            self._subnet = neutron.get_subnet(self._subnet,
                                              client=self.neutron_client)
        return self._subnet

    @property
    def has_subnet(self) -> bool:
        return self.subnet is not None

    def setup_stack_name(self) -> str:
        if self.stack_name is None:
            if self.has_subnet:
                self.stack_name = f"{self.fixture_name}-{self.subnet}"
            else:
                self.stack_name = f"{self.fixture_name}-{self.network}"
        return self.stack_name

    _port: typing.Optional[neutron.PortIdType] = None

    @property
    def port_details(self) -> neutron.PortType:
        if self._port is None:
            port_id = self.port_id
            if port_id is None:
                self._port = neutron.find_port(network_id=self.network_id,
                                               device_id=self.router_id,
                                               client=self.neutron_client)
            else:
                self._port = neutron.get_port(port_id,
                                              client=self.neutron_client)
        assert isinstance(self._port, dict)
        return self._port

    @property
    def neutron_client(self) -> neutron.NeutronClientType:
        if self._neutron_client is None:
            self._neutron_client = neutron.get_neutron_client()
        return self._neutron_client



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import socket
import typing
import weakref

import tobiko
from tobiko.shell.sh import _exception
from tobiko.shell.sh import _execute
from tobiko.shell import ssh


class HostnameError(tobiko.TobikoException):
    message = "Unable to get hostname from host: {error}"


HOSTNAMES_CACHE: typing.MutableMapping[typing.Optional[ssh.SSHClientFixture],
                                       str] = weakref.WeakKeyDictionary()


def get_hostname(ssh_client: ssh.SSHClientType = None,
                 cached=True,
                 **execute_params) -> str:
    ssh_client = ssh.ssh_client_fixture(ssh_client)
    if ssh_client is None:
        return socket.gethostname()

    if cached:
        try:
            hostname = HOSTNAMES_CACHE[ssh_client]
        except KeyError:
            pass
        else:
            return hostname

    hostname = ssh_hostname(ssh_client=ssh_client,
                            **execute_params)
    if cached:
        HOSTNAMES_CACHE[ssh_client] = hostname
    return hostname


def ssh_hostname(ssh_client: ssh.SSHClientFixture,
                 **execute_params) \
        -> str:
    tobiko.check_valid_type(ssh_client, ssh.SSHClientFixture)
    try:
        result = _execute.execute('hostname',
                                  ssh_client=ssh_client,
                                  **execute_params)
    except _exception.ShellCommandFailed as ex:
        raise HostnameError(error=ex.stderr) from ex

    line: str
    for line in result.stdout.splitlines():
        hostname = line.strip()
        if hostname:
            break
    else:
        raise HostnameError(error=f"Invalid result: '{result}'")
    return hostname



# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko import config
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import stacks


CONF = config.CONF


@keystone.skip_unless_has_keystone_credentials()
class NetworkTest(testtools.TestCase):
    """Tests network creation"""

    #: Stack of resources with a network with a gateway router
    stack = tobiko.required_fixture(stacks.NetworkStackFixture)

    def test_find_network_with_id(self):
        network = neutron.find_network(id=self.stack.network_id)
        self.assertEqual(self.stack.network_id, network['id'])

    def test_list_networks(self):
        networks = neutron.list_networks()
        network_ids = {n['id'] for n in networks}
        self.assertIn(self.stack.network_id, network_ids)

    def test_get_network_with_id(self):
        return self._test_get_network(network=self.stack.network_id)

    def test_get_network_with_details(self):
        return self._test_get_network(network=self.stack.network_details)

    def _test_get_network(self, network: neutron.NetworkIdType):
        observed = neutron.get_network(network)
        network_id = neutron.get_network_id(network)
        self.assertEqual(network_id, observed['id'])

    def test_create_network(self) -> neutron.NetworkType:
        network = neutron.create_network(name=self.id())
        self.assertIsInstance(network['id'], str)
        self.assertNotEqual('', network['id'])
        self.assertEqual(self.id(), network['name'])
        self.assertEqual(network['id'], neutron.get_network(network)['id'])
        return network

    def test_delete_network_with_id(self):
        network = self.test_create_network()
        neutron.delete_network(network['id'])
        self.assertRaises(neutron.NoSuchNetwork, neutron.get_network,
                          network['id'])

    def test_delete_network_with_details(self):
        network = self.test_create_network()
        neutron.delete_network(network)
        self.assertRaises(neutron.NoSuchNetwork, neutron.get_network,
                          network['id'])

    def test_delete_network_with_invalid_id(self):
        self.assertRaises(neutron.NoSuchNetwork,
                          neutron.delete_network, '<invalid-id>')

    def test_network_stack(self):
        network = neutron.get_network(self.stack.network_id)
        self.assertEqual(self.stack.port_security_enabled,
                         network['port_security_enabled'])
        if self.stack.has_ipv4:
            self.assertIn(self.stack.ipv4_subnet_id, network['subnets'])
        else:
            self.assertNotIn(self.stack.ipv4_subnet_id, network['subnets'])
        if self.stack.has_ipv6:
            self.assertIn(self.stack.ipv6_subnet_id, network['subnets'])
        else:
            self.assertNotIn(self.stack.ipv6_subnet_id, network['subnets'])



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.openstackclient import _exception
from tobiko.openstack.openstackclient import _client
from tobiko.openstack.openstackclient import _network
from tobiko.openstack.openstackclient import _port
from tobiko.openstack.openstackclient import _security_group
from tobiko.openstack.openstackclient import _security_group_rule
from tobiko.openstack.openstackclient import _subnet
from tobiko.openstack.openstackclient import _log


OSPCliError = _exception.OSPCliAuthError
OSPCliAuthError = _exception.OSPCliAuthError

execute = _client.execute

network_list = _network.network_list
network_show = _network.network_show
network_create = _network.network_create
network_delete = _network.network_delete
network_set = _network.network_set
network_unset = _network.network_unset

port_list = _port.port_list
port_show = _port.port_show
port_create = _port.port_create
port_delete = _port.port_delete
port_set = _port.port_set
port_unset = _port.port_unset

security_group_list = _security_group.security_group_list
security_group_show = _security_group.security_group_show
security_group_create = _security_group.security_group_create
security_group_delete = _security_group.security_group_delete
security_group_set = _security_group.security_group_set
security_group_unset = _security_group.security_group_unset

security_group_rule_list = _security_group_rule.security_group_rule_list
security_group_rule_show = _security_group_rule.security_group_rule_show
security_group_rule_create = _security_group_rule.security_group_rule_create
security_group_rule_delete = _security_group_rule.security_group_rule_delete

subnet_list = _subnet.subnet_list
subnet_show = _subnet.subnet_show
subnet_create = _subnet.subnet_create
subnet_delete = _subnet.subnet_delete
subnet_set = _subnet.subnet_set
subnet_unset = _subnet.subnet_unset

network_loggable_resources_list = _log.network_loggable_resources_list
network_log_create = _log.network_log_create
network_log_show = _log.network_log_show
network_log_list = _log.network_log_list
network_log_delete = _log.network_log_delete



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.openstackclient import _client


def subnet_list(*args, **kwargs):
    cmd = 'openstack subnet list {params}'
    kwargs['format'] = 'json'
    return _client.execute(cmd, *args, **kwargs)


def subnet_show(subnet, *args, **kwargs):
    cmd = f'openstack subnet show {{params}} {subnet}'
    kwargs['format'] = 'json'
    return _client.execute(cmd, *args, **kwargs)


def subnet_create(subnet_name, network_name, *args, **kwargs):
    cmd = f'openstack subnet create {{params}} --network '\
          f'{network_name} {subnet_name}'
    kwargs['format'] = 'json'
    return _client.execute(cmd, *args, **kwargs)


def subnet_delete(subnets, *args, **kwargs):
    cmd = f'openstack subnet delete {{params}} {" ".join(subnets)}'
    return _client.execute(cmd, *args, **kwargs)


def subnet_set(subnet, *args, **kwargs):
    cmd = f'openstack subnet set {{params}} {subnet}'
    return _client.execute(cmd, *args, **kwargs)


def subnet_unset(subnet, *args, **kwargs):
    cmd = f'openstack subnet unset {{params}} {subnet}'
    return _client.execute(cmd, *args, **kwargs)



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import
from __future__ import division

import io

import dpkt
from oslo_log import log

from tobiko.shell.tcpdump import _interface
from tobiko.shell.tcpdump import _parameters
from tobiko.shell import sh
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


def start_capture(capture_file: str,
                  interface: str = None,
                  capture_filter: str = None,
                  capture_timeout: int = None,
                  ssh_client: ssh.SSHClientType = None) \
        -> sh.ShellProcessFixture:

    parameters = _parameters.tcpdump_parameters(
        capture_file=capture_file,
        interface=interface,
        capture_filter=capture_filter,
        capture_timeout=capture_timeout)

    command = _interface.get_tcpdump_command(parameters)

    LOG.debug("Capturing traffic with the following command in background: %s",
              command)
    # when ssh_client is None, an ssh session is created on localhost
    process = sh.process(command=command,
                         ssh_client=ssh_client,
                         sudo=True)
    process.execute()
    return process


def stop_capture(process):
    process.kill(sudo=True)
    process.close()


def get_pcap(process,
             capture_file: str,
             ssh_client: ssh.SSHClientType = None) -> dpkt.pcap.Reader:
    stop_capture(process)

    stdout = sh.execute(f"cat '{capture_file}'",
                        ssh_client=ssh_client,
                        sudo=True,
                        decode_streams=False).stdout
    pcap = dpkt.pcap.Reader(io.BytesIO(stdout))
    return pcap



# Copyright 2023 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import abc
import collections
import typing

from oslo_log import log

import tobiko
from tobiko import config
from tobiko.openstack import keystone
from tobiko.openstack.neutron import _quota_set as neutron_quota
from tobiko.openstack.nova import _quota_set as nova_quota


LOG = log.getLogger(__name__)


class InvalidFixtureError(tobiko.TobikoException):
    message = "invalid fixture {name!r}"


class BaseResourceFixture(tobiko.SharedFixture):
    """Base class for fixtures both types: those which uses heat stacks and
    those which are not.
    """
    client: keystone.KeystoneClient = None
    project: typing.Optional[str] = None
    user: typing.Optional[str] = None

    def setup_fixture(self):
        self.setup_client()
        self.setup_project()
        self.setup_user()

    def setup_project(self):
        if self.project is None:
            self.project = keystone.get_project_id(session=self.session)

    def setup_user(self):
        if self.user is None:
            self.user = keystone.get_user_id(session=self.session)

    @property
    def session(self):
        return self.setup_client().session

    def setup_client(self) -> keystone.KeystoneClient:
        # returns self.client itself, if the keystone client was already
        # created; else, creates a new keystone client and returns it.
        self.client = keystone.keystone_client(self.client)
        return self.client

    def ensure_quota_limits(self):
        """Ensures quota limits before creating a new stack
        """
        try:
            self.ensure_neutron_quota_limits()
            self.ensure_nova_quota_limits()
        except (nova_quota.EnsureNovaQuotaLimitsError,
                neutron_quota.EnsureNeutronQuotaLimitsError) as ex:
            raise InvalidFixtureError(name=self.fixture_name) from ex

    def ensure_neutron_quota_limits(self):
        required_quota_set = self.neutron_required_quota_set
        if required_quota_set:
            neutron_quota.ensure_neutron_quota_limits(project=self.project,
                                                      **required_quota_set)

    def ensure_nova_quota_limits(self):
        required_quota_set = self.nova_required_quota_set
        if required_quota_set:
            nova_quota.ensure_nova_quota_limits(project=self.project,
                                                user=self.user,
                                                **required_quota_set)

    @property
    def neutron_required_quota_set(self) -> typing.Dict[str, int]:
        return collections.defaultdict(int)

    @property
    def nova_required_quota_set(self) -> typing.Dict[str, int]:
        return collections.defaultdict(int)


class ResourceFixture(BaseResourceFixture, abc.ABC):
    """Base class for fixtures intended to manage Openstack resources not
    created using Heat, but with openstacksdk or other component clients (such
    as neutronclient, novaclient, manilaclient, etc).
    Those resources will be shared by multiple tests and are not cleaned up
    when the tobiko execution ends.

    Child classes must define the following common attributes:
    _resource: the type of resource that the child class actually manages
    (server, network, loadbalancer, manila share, etc) initialized to None.
    _not_found_exception_tuple: (tuple of) type of exceptions that the
    resource_find method could raise in case of not finding the resource.

    Child classes must define any other attributes required by the
    resource_create, resource_delete and resource_find methods. Examples:
    prefixes and default_prefixlen are needed for subnet_pools; description and
    rules are needed for security_groups; etc.

    Child classes must define the resource_create, resource_delete and
    resource_find methods. In case of resource_create and resource_find, they
    should return an object with the type defined for self._resource.

    Child classes may optionally implement simple properties to access to
    resource_id and resource using a more representative name (these properties
    will simply return self.resource_id or self.resource, respectively).
    """

    name: typing.Optional[str] = None
    _resource: typing.Optional[object] = None
    _not_found_exception_tuple: typing.Type[Exception] = (Exception)

    @property
    def resource_id(self):
        if self.resource:
            return self._resource['id']

    @property
    def resource(self):
        if not self._resource:
            try:
                self._resource = self.resource_find()
            except self._not_found_exception_tuple:
                LOG.debug("%r not found.", self.name)
                self._resource = None
        return self._resource

    @abc.abstractmethod
    def resource_create(self):
        pass

    @abc.abstractmethod
    def resource_find(self):
        pass

    @abc.abstractmethod
    def resource_delete(self):
        pass

    def setup_fixture(self):
        super().setup_fixture()
        self.name = self.fixture_name
        if config.get_bool_env('TOBIKO_PREVENT_CREATE'):
            LOG.debug("%r should have been already created: %r",
                      self.name,
                      self.resource)
        else:
            self.try_create_resource()

        if self.resource is None:
            tobiko.fail("%r not found!", self.name)
        else:
            tobiko.addme_to_shared_resource(__name__, self.name)

    def try_create_resource(self):
        # Ensure quota limits are OK just in time before start creating
        # a new stack
        self.ensure_quota_limits()
        if not self.resource:
            self._resource = self.resource_create()

    def cleanup_fixture(self):
        tests_using_resource = tobiko.removeme_from_shared_resource(__name__,
                                                                    self.name)
        if len(tests_using_resource) == 0:
            self._cleanup_resource()
        else:
            LOG.info(f'{self.name} ResourceFixture not deleted because some '
                     f'tests are still using it: {tests_using_resource}')

    def _cleanup_resource(self):
        resource_id = self.resource_id
        if resource_id:
            LOG.debug('Deleting %r (%r)...', self.name, resource_id)
            self.resource_delete()
            LOG.debug('%r (%r) deleted.', self.name, resource_id)
            self._resource = None



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import itertools

from oslo_config import cfg

GROUP_NAME = "topology"
OPTIONS = [
    cfg.ListOpt('nodes',
                default=None,
                help="List of hostname nodes"),
    cfg.StrOpt('key_file',
               default=None,
               help="Default SSH key to login to cloud nodes"),
    cfg.StrOpt('username',
               default=None,
               help="Default username for SSH login"),
    cfg.StrOpt('port',
               default=None,
               help="Default port for SSH login"),
    cfg.StrOpt('ip_version',
               default=None,
               choices=['', '4', '6'],
               help="Limit connectivity to cloud to IPv4 o IPv6"),
    cfg.StrOpt('log_datetime_pattern',
               default=r"(\d{4}-\d{2}-\d{2} [0-9:.]+) .+",
               help="Regex to be used to parse date and time from "
                    "the OpenStack services' log lines. "),
]


def register_tobiko_options(conf):
    conf.register_opts(group=cfg.OptGroup(GROUP_NAME), opts=OPTIONS)


def list_options():
    return [(GROUP_NAME, itertools.chain(OPTIONS))]
